//////////////////////////////////////////////////////////////////////////
// Busy soft // LzxList 02 (depacker & dumper & analyzer) // 30.08.2021 //
//////////////////////////////////////////////////////////////////////////

#define DBG 0

///////////////////////////
// Types of compressions //
///////////////////////////

// Systems for coding unpacked data and sekvence lengths //

#define COMBLX 1	// Simple block compression - elias-gama length for both sekvence and unpacked block (real sequence 1+ length supports compress simple bytes)
#define COMBLC 2	// Simple block compression - elias-gama length for both sekvence and unpacked block (real sequence 2+ length has width more 1 bit: EG+1)
#define COMZX9 3	// Similar ZX7 compression - 1 bit for each unpacked byte and 1 bit + elias-gamma length for sekvence 1+ bytes (supports packing of simple bytes)
#define COMZX8 4	// Similar ZX7 compression - 1 bit for each unpacked byte and 1 bit + elias-gamma length for sekvence 2+ bytes (length EG+1 bit)
#define COMBS2 5	// Bitstream 1:unpack byte, 01:sek 2 bytes, 001:sek 3 bytes, 0001:sek 4+ bytes, 00000:1 unpack + rep offset, 00000:unpacked block 10+ bytes
#define COMBX1 6	// Bitstream 1:unpack, 01:sek2, 001:sek3, 0001:copied, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMBX2 7	// Bitstream 1:unpack, 01:sek2, 001:copied, 0001:sek3, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMBX3 8	// Bitstream 1:unpack, 01:copied, 001:sek2, 0001:sek3, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMSX1 9	// Bitstream 1:sek 2+, 01:unpack, 001:copied, 0001:unpack + rep offset, 00001:copied + rep offset, 00000:unpack block 5+ bytes

// Systems for coding offsets in sequences //

#define POSOF1 1	// One fixed offset used for all sekvences
#define POSOF2 2	// Two fixed offsets (similar than ZX7)
#define POSOF3 3	// Three fixed step offsets (1,2,3 or 2,4,6 or 3,6,9 or 4,8,12 or 5,10,15)
#define POSOF4 4	// Four fixed step offsets (1,2,3,4 or 2,4,6,8 or 3,6,9,12 or 4,8,12,16 bits)
//efine POSxxx 5	// Not used (was OV0 but too complex and not used in comparation test)
#define POSOV1 6	// Simple variable elias-gama coding offset
#define POSOV2 7	// One fixed offset and one variable offset
#define POSOV3 8	// Two fixed offsets A,B and one variable elias-gama offset EG+B bits and +A addition
#define POSOV4 9	// Three fixed step offsets (1,2,3 or 2,4,6 or 3,6,9 or 4,8,12) and one variable offset

// Compression characteristics //

#define nie 0	// No
#define ANO 1	// Yes
// ............................   -     1     2     3     4     5     6     7     8     9     -		// Compression identification
const char *compress_name[] = { ":::","BLX","BLC","ZX9","ZX8","BS2","BX1","BX2","BX3","SX1","###"};	// Compression name identifier and abbreviation
const char ena_first_byte[] = {   0,   nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,   0 };	// Compressed data begins by one uncompressed byte without any overhead
const char ena_rep_offset[] = {   0,   nie,  nie,  nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,   0 };	// Enable re-using the same offset value from previous sequence
const char max_1byte_sekv[] = {   0,     6,  nie,    6,  nie,  nie,    4,    5,    6,    6,   0 };	// Maximal bits in offset of one copied byte (0 = no packing single bytes)

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

// int _CRT_glob = 0; //

#define MAXLEN 80000

int bits;
int data_len;
int over_len;
int overhead;

int packed_size;
int packed_index;
int unpack_index;
int last_offset;

int reverse_order;
int compress_type;
int compress_posi;
int offset_byte;
int offset1bits;
int offset2bits;

int user_rev_order;
int user_com_type;
int user_com_posi;
int user_offs_byte;
int user_offs1bits;
int user_offs2bits;

int show_pck_zac;
int show_pck_end;
int show_unp_zac;
int show_unp_end;
int show_unp_len;

char buftmp[1024];
char  unpack_name[1024];
char *packed_name = NULL;
unsigned char packed_data[MAXLEN];
unsigned char unpack_data[MAXLEN];

int limits[32];			// Values for distribution of following mapsek statistics
int mapsek[32][32][2];	// Sekvence map of pack model - offset x length x own/rep
int stat[3][2][65536];	// Counters of unpacked blocks and sekvences: stat[what][reused][length]

#define UNPLEN 0	// Index for lengths of unpacked blocks
#define SEKLEN 1	// Index for lengths of sekvences
#define SEKOFF 2	// Index for offsets of sekvences

////////////////////////////////////////////////////////////////////////////////

void DisplayCompression()
{
	if ((unsigned)compress_type > 9) printf("???");
	else printf(compress_name[compress_type]);

	printf(" %c ", max_1byte_sekv[compress_type] ? offset_byte | '0' : '-');
	int z = 0;

	switch (compress_posi)
	{
	case POSOF1: z = printf("OF1 fixed offset %2u", offset1bits); break;
	case POSOF2: z = printf("OF2 offsets %2u %u", offset1bits, offset2bits); break;
	case POSOF3: z = printf("OF3 offs %u %u %u", offset1bits, 2 * offset1bits, 3 * offset1bits); break;
	case POSOF4: z = printf("OF4 offs %u %u %u %u", offset1bits, 2 * offset1bits, 3 * offset1bits, 4 * offset1bits); break;
	case POSOV1: z = printf("OV1 variable offset"); break;
	case POSOV2: z = printf("OV2 offsets %2u var", offset1bits); break;
	case POSOV3: z = printf("OV3 offs %2u %2u var", offset1bits, offset2bits); break;
	case POSOV4: z = printf("OV4 offs %u %u %u var", offset1bits, 2 * offset1bits, 3 * offset1bits); break;
	default:	 z = printf("(unknown)");
	}
	printf(" %s ", "...................." + z);
}

int GetNumberOffsets()
{
	switch (compress_posi)
	{
	case POSOF2:
	case POSOV3: return 2;
	case POSOF3:
	case POSOF4:
	case POSOV4:
	case POSOF1:
	case POSOV2: return 1;
	case POSOV1:
	default: return 0;
	}
}

int GetByte(void)
{
	// Read one byte from packed data

	if (DBG > 2) printf(" [%02X] ", packed_data[packed_index]);
	return packed_data[packed_index++];
}

int GetBit(void)
{
	// Read one bit from bitstream

	bits <<= 1;
	if (!(bits & 0xFF))
	{
		bits = (GetByte() << 1) | 1;
		over_len++;
	}
	overhead++;
	if (DBG > 1) printf(" %c ", ((bits >> 8) & 0x01) | '0');
	return (bits >> 8) & 0x01;
}

int GetValue(int bitwide)
{
	// Read direct value of bitwide bits

	if (DBG > 1) printf(" (ValBit:%u) ", bitwide);
	int value = 0;
	for (int i = 0; i < bitwide; i++) value = (value << 1) | GetBit();
	if (DBG > 1) printf(" (ValRet:%u) ", value);
	return value;
}

int GetEliasGama(int bitwide)
{
	// Read dynamically stored number in format:  <n*zero bits> 1 <value n+bitwide bits>

	if (DBG > 1) printf(" (EGbit:%u) ", bitwide);
	while (!GetBit() && (bitwide < 18)) bitwide++;
	if (bitwide > 17) { printf("\nError in %s: Too many %u+ zero bits\n", packed_name, bitwide); return 0; }
	if (bitwide > 15) { if (DBG > 1) printf(" (End mark) "); return 0; }
	if (DBG > 1) printf(" (B:%u) ", bitwide);
	int value = 1;
	for (int i = 0; i < bitwide; i++) value = (value << 1) | GetBit();
	if (DBG > 1) printf(" (EGval:%u) ", value);
	return value;
}

int GetBlockParams(int *resoff, int *reslen, unsigned *resbyt)
{
	// Read and evaluate one data block specification //
	// It depends to type of compression and offset ! //

	// Output:
	//
	// Return != 0 ... Error occured (length not valid)
	// Return = 0 .... OK, see following length meaning:
	//
	// length = 0 .... end mark (offset not valid)
	// length > 0 .... length of unpacked data or sekvence, see following offset meaning:
	//
	// offset = 0 .... unpacked data
	// offset > 0 .... packed sekvence with own offset
	// offset < 0 .... one copied byte and sekvence with reused offset
	// offset = FFFF ... unpacked byte and sekvence with reused offset (byte in resbyt)

	int packed = 0;
	int length = 0;
	int cc = compress_type;

	if (DBG > 1) printf(" (Blk) ");

	// Get pack/unpack flag and length of data block from COMxxx //

	switch (compress_type)
	{
	case COMBLX:
		length = GetEliasGama(0);
		packed = !GetBit();
		break;

	case COMBLC:
		packed = !GetBit();
		length = GetEliasGama(packed);
		break;

	case COMZX8:
		length = 1;
		packed = !GetBit();
		if (packed) length = GetEliasGama(1);
		break;

	case COMZX9:
		length = 1;
		packed = !GetBit();
		if (packed) length = GetEliasGama(0);
		break;

	case COMBS2:
		packed = 0;
		if (GetBit()) { length = 1; break; } // 1
		packed = 1;
		if (GetBit()) { length = 2; break; } // 01
		if (GetBit()) { length = 3; break; } // 001
		if (GetBit()) { length = GetEliasGama(2); break; } // 0001
		if (GetBit())
		{
			// 00001  Unpack + rep offset
			if (DBG > 1) printf(" (UnpRep) ");
			*resoff = 0xFFFF;
			*resbyt = GetByte();
			*reslen = GetEliasGama(1);
			return 0;
		}
		length = GetEliasGama(3) + 2; // 00000
		if (length == 2) length = 0;
		packed = 0;
		break;

	case COMBX1:
	case COMBX2:
	case COMBX3:
		packed = 0;
		if (GetBit()) { length = 1; break; } // 1 One unpacked byte
		packed = 1;
		if (GetBit()) { length = cc == COMBX1 ? 2 : cc == COMBX2 ? 2 : /* COMBX3 */ 1; break; } // 01
		if (GetBit()) { length = cc == COMBX1 ? 3 : cc == COMBX2 ? 1 : /* COMBX3 */ 2; break; } // 001
		if (GetBit()) { length = cc == COMBX1 ? 1 : cc == COMBX2 ? 3 : /* COMBX3 */ 3; break; } // 0001
		if (GetBit()) { length = GetEliasGama(2); break; } // 00001 Regular sekvence 4+ bytes
		if (GetBit())
		{
			// 000001 Unpack + rep offset
			if (DBG > 1) printf(" (UnpRep) ");
			*resoff = 0xFFFF;
			*resbyt = GetByte();
			*reslen = GetEliasGama(1);
			return 0;
		}
		if (GetBit())
		{
			// 0000001 One copied + rep offset
			if (DBG > 1) printf(" (CopRep) ");
			*resoff = -1 - GetValue(offset_byte);
			*reslen = GetEliasGama(1);
			return 0;
		}
		length = GetEliasGama(3) + 4; // 0000000 Block 12+ unpacked bytes
		packed = 0;
		break;

	case COMSX1:
		packed = 1;
		if (GetBit()) { length = GetEliasGama(1); break; } // 1 Sekvence 2+ bytes
		packed = 0;
		if (GetBit()) { length = 1; break; } // 01 One unpacked byte
		packed = 1;
		if (GetBit()) { length = 1; break; } // 001 One copied byte
		if (GetBit())
		{
			// 0001 Unpack + rep offset
			if (DBG > 1) printf(" (UnpRep) ");
			*resoff = 0xFFFF;
			*resbyt = GetByte();
			*reslen = GetEliasGama(1);
			return 0;
		}
		if (GetBit())
		{
			// 00001 One copied + rep offset
			if (DBG > 1) printf(" (CopRep) ");
			*resoff = -1 - GetValue(offset_byte);
			*reslen = GetEliasGama(1);
			return 0;
		}
		length = GetEliasGama(2) + 1; // 00000 Block 5+ unpacked bytes
		packed = 0;
		break;

	default:
		printf("\nError in %s: Unknown compression type %u\n", packed_name, compress_type);
		return 1;
	}

	*resoff = 0;
	*reslen = length;

	if (!length) return 0; // End mark //
	if (!packed) return 0; // Unpacked data //

	if (DBG > 1) printf(" (Ofs) ");

	if (length == 1) // Get offset for one copied byte = allways fixed //
	{
		*resoff = GetValue(offset_byte) + 1;
		return 0;
	}

	// Get regular offset POSxxx //

	int width = 999999;
	int offset = 999999;

	switch (compress_posi)
	{
	case POSOF1:
		offset = GetValue(offset1bits) + 1;
		break;

	case POSOF2:
		width = GetBit();
		offset = GetValue(width ? offset1bits : offset2bits) + 1;
		if (!width) offset += 1 << offset1bits;
		break;

	case POSOF3:
		width = 1;
		if (GetBit()) width = 2 + GetBit();
		offset = GetValue(offset1bits * (width));
		for (int adds = width - 1; adds >= 0; adds--) offset += 1 << (adds * offset1bits);
		break;

	case POSOF4:
		width = GetBit() << 1;
		width |= GetBit();
		offset = GetValue(offset1bits * (width + 1));
		for (int adds = width; adds >= 0; adds--) offset += 1 << (adds * offset1bits);
		break;

	case POSOV1:
		offset = GetEliasGama(0);
		break;

	case POSOV2:
		if (!GetBit()) { offset = GetValue(offset1bits) + 1; break; }	// 0 <short-offset>
		offset = GetEliasGama(offset1bits) + 1;	// 1 Elias-gama offset
		break;

	case POSOV3:
		if (!GetBit()) { offset = GetValue(offset1bits) + 1; break; }	// 0 <short-offset>
		if (!GetBit()) { offset = GetValue(offset2bits) + 1 + (1 << offset1bits); break; }	// 10 <long-offset>
		offset = GetEliasGama(offset2bits) + (1 << offset1bits) + 1;	// 11 Elias-gama offset
		break;

	case POSOV4:
		offset = 1;
		width = GetBit() << 1;
		width |= GetBit();
		for (int adds = width; adds > 0; adds--) offset += 1 << (adds * offset1bits);
		if (width < 3) offset += GetValue(offset1bits * (width + 1));
		else offset = GetEliasGama(3 * offset1bits) + 1 + (1 << offset1bits) + (1 << (2 * offset1bits));
		break;

	default:
		printf("\nError in %s: Unknown offset coding %u\n", packed_name, compress_posi);
		return 1;
	}

	if (!offset)
	{
		printf("\nError in %s: Zero offset\n", packed_name);
		return 1;
	}

	*resoff = offset;
	return 0;
}

void PutCompressType(void)
{
	// Display compression type -tnXYcNoAoB //

	printf("-t%c%c%c", reverse_order ? 'r' : 'n', (compress_type & 0x0F) | '0', (compress_posi & 0x0F) | '0');
	if (compress_type <= 9 && max_1byte_sekv[compress_type]) printf("c%u", offset_byte);
	int numofs = GetNumberOffsets();
	if (numofs)
	{
		printf("o%u", offset1bits);
		if (numofs > 1) printf("o%u", offset2bits);
	}
	putchar('\n');
}

int GetCompressType(char * txt, char *where)
{
	// Evaluate compress type -tnXYcNoAoB from file name or command line parameter //

	if (DBG) printf("Evaluate: %s\nIn place: %s\n", txt, where);

	reverse_order = -1;
	compress_type = 0;
	compress_posi = 0;
	offset_byte = 0;
	offset1bits = 0;
	offset2bits = 0;

	switch (*txt | 0x20)
	{
	case 'n': reverse_order = 0; break;
	case 'r': reverse_order = 1; break;
	default: printf("Error: Compression direction -tr or -tn not selected %s\n", where); return -1;
	}
	if (DBG) printf("Reverse: %c => %u\n", *txt, reverse_order);
	txt++;
	char znak = *txt;
	if (!znak) { printf("Error: Compression type -tXY not selected %s\n", where); return -1; }
	compress_type = znak - '0';
	if (DBG) printf("Compres: %c => %d\n", znak, compress_type);
	if (compress_type < 1 || compress_type > 9) { printf("Error: Unknown compression type '%c' %s\n", znak, where); return -1; }
	txt++;
	znak = *txt;
	if (!znak) { printf("Error: Coding offset -tXX not selected %s\n", where); return -1; }
	compress_posi = znak - '0';
	int offset_number = GetNumberOffsets();
	if (DBG) printf("Offsets: %c => %d  Num: %d\n", znak, compress_posi, offset_number);
	if (offset_number < 0) { printf("Error: Unknown coding offset '%c' %s\n", znak, where); return -1; }

	if (max_1byte_sekv[compress_type])
	{
		txt++;
		if ((*txt | 0x20) != 'c') { printf("Error: Cannot determine one copied byte offset %s\n", where); return -1; }
		txt++;
		znak = *txt;
		if (!znak) { printf("Error: One copied byte offset -cX not selected %s\n", where); return -1; }
		offset_byte = znak - '0';
		if (DBG) printf("One copied byte offset: %c => %d\n", znak, offset_byte);
		if (offset_byte < 1 || offset_byte > 9) { printf("Error: Unknow one copied byte offset '%c' %s\n", znak, where); return -1; }
	}
	if (offset_number)
	{
		txt++;
		if ((*txt | 0x20) != 'o') { printf("Error: Cannot determine 1st offsets %s\n", where); return -1; }
		while (*txt)
		{
			txt++;
			znak = *txt;
			if (znak <'0' || znak > '9') break;
			offset1bits = 10 * offset1bits + znak - '0';
		}
		if (offset_number == 2)
		{
			if ((*txt | 0x20) != 'o') { printf("Error: Cannot determine 2nd offsets %s\n", where); return -1; }
			while (*txt)
			{
				txt++;
				znak = *txt;
				if (znak <'0' || znak > '9') break;
				offset2bits = 10 * offset2bits + znak - '0';
			}
		}
	}
	if (znak && !*where) { printf("Error: Unexpected characters in type selection: %s\n", txt); return -1; }
	if (offset1bits > 16 || offset2bits > 16) { printf("Error: Offset overflows maximal value 16 %s\n", where); return -1; }
	if (offset2bits && (offset2bits <= offset1bits || offset2bits <= 2))
	{
		printf("Error: 2nd offset %u must be bigger than %s %s\n", offset2bits, offset2bits <= 2 ? "2" : "1st offset", where); return -1;
	}
	return 0;
}

void ShowHeader(int pckzac, int unpzac)
{
	int pck = reverse_order ? packed_size - pckzac : pckzac;
	printf("Pack: %5d %04X   Unpack: %5u %04X   ", pck, pck, unpzac, unpzac);
}

void ShowFooter(int offset, int length)
{
	int i;
	printf("Length: %-5u  Data:", length);
	for (i = 0; (i < length) && (i < 16); i++)
		if (offset >= 0 && offset < MAXLEN)
			printf(" %02X", unpack_data[offset+i]);
		else
			printf(" XX");
	if (i < length) printf(" ...");
	putchar('\n');
}

void AddSekvence(int offset, int length, int reused)
{
	// Increment counters in statistic map array //

	int idxoff = offset; if (idxoff > 16) { idxoff = 17; while (offset > limits[idxoff]) idxoff++; }
	int idxlen = length; if (idxlen > 16) { idxlen = 17; while (length > limits[idxlen]) idxlen++; }
	mapsek[idxoff][idxlen][reused]++;
}

int main(int number_of_params, char **parameters)
{
	if (number_of_params < 2)
	{
		puts(
			"\nBusy soft: Universal LZX Lister and Depacker 02\n\n"
			"Usage: LzxList <options> file1 file2 ...\n\n"
			"Options:\n"
			" -l ............ Output listing of packed files to stdout\n"
			" -s ............ Output detailed statistics of packed files to stdout\n"
			" -d ............ Depack and output files with extension .out (or set by -e)\n"
			" -u ............ Same as -d but with compression info -tXX.. removed from the filename\n"
			" -e <ext> ...... Set output file extension for depacked output (default extension: out)\n"
			" -o <file> ..... Set file name for depacked data (default: input file + extension .out)\n"
			" -tnXYcNoAoB ... Select compression manually (if no compression type -tXY.. in filename)\n"
			" -trXYcNoAoB ... The 'r' instead of 'n' forces reverse direction of depacking\n\n"
			"You can use wildcards in filenames to process more files at once.\n"
		);
		exit(1);
	}
	putchar('\n');

	user_rev_order = -1;
	user_com_type = 0;
	user_com_posi = 0;
	user_offs_byte = 0;
	user_offs1bits = 0;
	user_offs2bits = 0;

	char *extension = NULL;
	char *out_name = NULL;
	int make_list = 0;
	int make_stat = 0;
	int make_unpack = 0;
	int some_success = 0;
	int number_of_files = 0;

	for (int param = 1; param < number_of_params; param++)
	{
		if (DBG) printf("Pass 1 param %u: %s\n", param, parameters[param]);
		if (parameters[param][0] == '-')
		{
			char * xx = parameters[param] + 1;
			switch (*xx | 0x20)
			{
			case 'l':
				if (DBG) puts("Option -L found");
				make_list = 1;
				break;
			case 's':
				if (DBG) puts("Option -S found");
				make_stat = 1;
				break;
			case 'd':
				if (DBG) puts("Option -D found");
				if (make_unpack == 2) printf("Error: You can not use -D and -U together.\n");
				make_unpack = 1;
				break;
			case 'u':
				if (DBG) puts("Option -U found");
				if (make_unpack == 1) printf("Error: You can not use -D and -U together.\n");
				make_unpack = 2;
				break;
			case 'e':
				xx++;
				if (!*xx)
				{
					if (param < number_of_params - 1) { param++; xx = parameters[param]; }
					else { puts("Error: No extension for -e option"); exit(1); }
				}
				extension = xx;
				if (DBG) printf("Option -E set extension to %s\n", *extension ? extension : "(none)");
				break;
			case 'o':
				xx++;
				if (!*xx)
				{
					if (param < number_of_params - 1) { param++; xx = parameters[param]; }
					else { puts("Error: No filename for -o option"); exit(1); }
				}
				out_name = xx;
				if (DBG) printf("Option -O set output name to %s\n", out_name);
				make_unpack = 1;
				break;
			case 't':
				if (GetCompressType(xx + 1, (char*)"in cmdline")) break;
				user_rev_order = reverse_order;
				user_com_type = compress_type;
				user_com_posi = compress_posi;
				user_offs_byte = offset_byte;
				user_offs1bits = offset1bits;
				user_offs2bits = offset2bits;
				printf("Selected compression type: ");
				PutCompressType();
				break;
			default:
				printf("Error: Invalid option %s\n", parameters[param]);
			}
		}
		else number_of_files++;
	}

	if (!number_of_files) { printf("Error: No input files\n"); exit(1); }

	int i;
	for (i = 0; i < 32; i++) limits[i] = i <= 16 ? i : 1 << (i - 12);
	if (DBG) { printf("Limits:"); for (i = 0; i < 32; i++) printf(" %u", limits[i]); puts("\n"); }

	for (int param = 1; param < number_of_params; param++)
	{
		if (DBG) printf("Pass 2 param %u: %s\n", param, parameters[param]);
		if (parameters[param][0] == '-')
		{
			if (((parameters[param][1] | 0x20) == 'e') && !parameters[param][2]) param++;
			if (((parameters[param][1] | 0x20) == 'o') && !parameters[param][2]) param++;
			continue;
		}

		packed_name = parameters[param];
		strcpy(unpack_name, packed_name);
		int name_length = strlen(unpack_name);
		int end_of_name = name_length;
		if (end_of_name > 5)
		{
			char * dotpos = strrchr(unpack_name + end_of_name - 5, '.');
			if (dotpos) { *dotpos = 0; end_of_name = dotpos - unpack_name; }
		}

		////////////////////////////////////////////

		reverse_order = -1;
		compress_type = 0;
		compress_posi = 0;
		offset_byte = 0;
		offset1bits = 0;
		offset2bits = 0;

		int begin_type = end_of_name;
		int idx;
		for (idx = end_of_name - 1; idx > 0; idx--)
			if (packed_name[idx] == '-' && (packed_name[idx + 1] | 0x20) == 't') break;

		if (DBG)
		{
			printf("Compress type: ");
			for (int i = idx; i < end_of_name; i++) putchar(packed_name[i]);
			putchar('\n');
		}

		begin_type = idx;
		sprintf(buftmp, "in file name %s", packed_name);
		if (GetCompressType(packed_name + idx + 2, buftmp)) continue;

		if (reverse_order == -1) reverse_order = user_rev_order;
		if (!compress_type) compress_type = user_com_type;
		if (!compress_posi) compress_posi = user_com_posi;
		if (!offset_byte) offset_byte = user_offs_byte;
		if (!offset1bits) offset1bits = user_offs1bits;
		if (!offset2bits) offset2bits = user_offs2bits;

		if (!compress_type || compress_type > 9) { printf("Error: Cannot determine compress type for file %s\n", packed_name); continue; }

		if (DBG)
		{
			printf("File cmpress type: "); PutCompressType();
			printf("First unpack byte: %d\n", ena_first_byte[compress_type]);
		}

		////////////////////////////////////////////

		if (make_unpack == 2) end_of_name = begin_type;
		unpack_name[end_of_name] = '.';
		strcpy(unpack_name + end_of_name + 1, "out");

		if (DBG)
		{
			printf("List: %u  Depack: %u\nOutput file: %s\n", make_list, make_unpack, unpack_name);
			printf("Compression: "); DisplayCompression();
			printf("\nCom:%u  Pos:%u  Byte:%u  Off1:%u  Off2:%u\nInput file: %s\n\n",
				compress_type, compress_posi, offset_byte, offset1bits, offset2bits, packed_name);
		}

		FILE *packed_file = fopen(packed_name, "rb"); if (!packed_file) { perror(packed_name); continue; }
		packed_size = fread(packed_data, 1, MAXLEN, packed_file);
		fclose(packed_file);
		memset(packed_data + packed_size, 0x00, 16);

		if (make_list)
		{
			printf(
				"List of file ... %s  (%u bytes)\n"
				"Compression .... ",
				packed_name, packed_size);
			DisplayCompression();
			puts("\n");
		}

		if (reverse_order)
			for (int i = (packed_size >> 1) - 1; i >= 0; i--)
			{
				unsigned char tmp = packed_data[i];
				packed_data[i] = packed_data[packed_size - i - 1];
				packed_data[packed_size - i - 1] = tmp;
			}

		memset(stat, 0x00, sizeof(stat));
		memset(mapsek, 0x00, sizeof(mapsek));

		bits = 0x80;
		overhead = 0;
		packed_index = 0;
		unpack_index = 0;

		int packed_count = 0;
		int unpack_count = 0;

		if (ena_first_byte[compress_type])
		{
			unpack_data[unpack_index++] = packed_data[packed_index++];
			unpack_count++;
			if (DBG || make_list)
			{
				ShowHeader(0, 0);
				printf("First standalone byte ........... ");
				ShowFooter(0, 1);
			}
		}

		show_unp_len = 0;

		int err = 0;
		int numone = 0;
		int numsek = 0;
		int blok_packed = 0;
		int blok_unpack = 0;
		int previous_offset = 0;
		int previous_unpack = 0;
		int offset, length;
		unsigned unbyte;

		while (packed_index < packed_size)
		{
			blok_packed = packed_index;
			blok_unpack = unpack_index;

			if (GetBlockParams(&offset, &length, &unbyte)) { err++; break; }
			if (DBG > 1) putchar('\n');

			if ((!length || offset) && show_unp_len && (DBG || make_list || make_stat))
			{
				previous_unpack = show_unp_len;
				stat[UNPLEN][0][previous_unpack]++;

				if (make_list)
				{
					ShowHeader(show_pck_zac, show_unp_zac);
					printf("Unpacked block .................. ");
					ShowFooter(show_unp_zac, show_unp_len);
				}
				show_unp_len = 0;
			}

			if (!length)
			{
				if (DBG || make_list)
				{
					ShowHeader(blok_packed, unpack_index);
					printf("End mark\n");
					ShowHeader(packed_index, unpack_index);
					printf("End of file\n");
				}
				break;
			}
			if (length + unpack_index > MAXLEN)
			{
				printf("\nError in %s at +%u: Unpack %u + length %u > buffer %u\n",
					packed_name, packed_index, unpack_index, length, MAXLEN); err++; break;
			}

			if (offset)
			{
				if (offset == 0xFFFF)	// One unpack byte and sekvence with reused offset //
				{
					stat[SEKLEN][1][length]++;
					stat[SEKOFF][1][previous_offset]++;
					if (previous_unpack)
					{
						stat[UNPLEN][0][previous_unpack]--;
						stat[UNPLEN][1][previous_unpack + 1]++;
					}
					else
					{
						stat[UNPLEN][1][1]++;
					}
					AddSekvence(previous_offset, length, 1);

					unpack_data[unpack_index] = unbyte;
					unpack_index++;
					unpack_count++;
					numsek++;
					if (DBG || make_list)
					{
						ShowHeader(blok_packed, blok_unpack);
						printf("One unpacked byte ............... ");
						ShowFooter(unpack_index - 1, 1);
						ShowHeader(blok_packed, blok_unpack + 1);
						printf("Sekvence .. Reused Offset: %-5u  ", previous_offset);
						// ShowFooter(unpack_index - previous_offset, length);
					}
					offset = previous_offset;
				}
				else if (offset < 0)	// One copied byte and sekvence with reused offset //
				{
					offset = -offset;

					stat[SEKLEN][0][1]++;
					stat[SEKLEN][1][length]++;
					stat[SEKOFF][0][offset]++;
					stat[SEKOFF][1][previous_offset]++;
					AddSekvence(offset, 1, 0);
					AddSekvence(previous_offset, length, 1);

					unpack_data[unpack_index] = unpack_data[unpack_index - offset];
					unpack_index++;
					packed_count++;
					numsek++;
					numone++;
					if (DBG || make_list)
					{
						ShowHeader(blok_packed, blok_unpack);
						printf("One copied byte .. Offset: %-5u  ", offset);
						ShowFooter(unpack_index - 1, 1);
						ShowHeader(blok_packed, blok_unpack + 1);
						printf("Sekvence .. Reused Offset: %-5u  ", previous_offset);
					}
					if (unpack_index < offset)
					{
						printf("\nError in %s at +%u: Unpack %u - one byte offset %u < 0\n", packed_name, packed_index, unpack_index, offset);
						err++;
						break;
					}
					offset = previous_offset;
				}
				else
				{
					stat[SEKLEN][0][length]++;
					stat[SEKOFF][0][offset]++;
					AddSekvence(offset, length, 0);

					if (length == 1)
					{
						numone++;
					}
					else
					{
						numsek++;
						previous_offset = offset;
					}

					if (DBG || make_list)
					{
						ShowHeader(blok_packed, blok_unpack);
						if (length == 1)
							printf("One copied byte .. Offset: %-5u  ", offset);
						else
							printf("Sekvence ......... Offset: %-5u  ", offset);
					}
				}

				if (unpack_index < offset)
				{
					printf("\nError in %s at +%u: Unpack %u - offset %u < 0\n", packed_name, unpack_index, unpack_index, offset);
					err++;
					break;
				}
				if (unpack_index + length > MAXLEN)
				{
					printf("\nError in %s at +%u: Unpack %u + length %u > buffer %u\n", packed_name, unpack_index, unpack_index, length, MAXLEN);
					err++;
					break;
				}

				// memcpy(unpack_data + unpack_index, unpack_data + unpack_index - offset, length);
				for (i = 0; i < length; i++) unpack_data[unpack_index + i] = unpack_data[unpack_index - offset + i];

				if (DBG || make_list) ShowFooter(unpack_index, length);

				if (DBG > 2) printf("Copy sekvence: From: %u %04X  To: %u %04X  Length: %u  Offset: %-5u\n",
						unpack_index - offset, unpack_index - offset, unpack_index, unpack_index, length, offset);

				unpack_index += length;
				packed_count += length;

				previous_unpack = 0;
			}
			else
			{
				if (!show_unp_len)
				{
					show_pck_zac = blok_packed;
					show_unp_zac = blok_unpack;
					if (DBG > 2) printf("Set unpack block addresses: Pack: +%u  Unpack: +%u\n", show_pck_zac, show_unp_zac);
				}

				if (DBG)
				{
					int i;
					printf("Temporary unpack block   at %-5u Len:%-5u Data:", blok_unpack, length);
					for (i = 0; (i < length) && (i < 16) && (i + packed_index < packed_size); i++) printf(" %02X", packed_data[packed_index + i]);
					if (i < length) printf(" ...");
					putchar('\n');
				}

				if (length + packed_index > packed_size)
				{
					printf("\nError in %s at +%u: Source index + length %u > file size %u\n",
						packed_name, packed_index, length, packed_size); err++;
					break;
				}

				memcpy(unpack_data + unpack_index, packed_data + packed_index, length);
				// for (i = 0; i < length; i++) unpack_data[unpack_index + i], packed_data[unpack_index + i];

				unpack_index += length;
				packed_index += length;
				unpack_count += length;
				show_unp_len += length;

				show_pck_end = packed_index - 1;
				show_unp_end = packed_index - 1;
			}
		}
		if (err) continue;
		if (DBG || make_list) putchar('\n');

		if (!make_list && !make_stat)
			printf("%-32s  %5u => %-5u  Sek: %-4u  Pck: %-5u  NoPck: %-5u  Overhead: %-5u (%u bits)\n",
				packed_name, packed_index, unpack_index, numsek + numone, packed_count, unpack_count, (overhead + 7) >> 3, overhead);
		else
			printf(
				"  Total packed length ...... %u (%u bits)\n"
				"  Total unpack length ...... %u\n"
				"  Number of sekvences ...... %u\n"
				"  Number of copied bytes ... %u\n"
				"  Number of packed bytes ... %u\n"
				"  Number of unpack bytes ... %u\n"
				"  Compresion overhead ...... %u (%u bits)\n\n",
				packed_index, overhead + (unpack_count << 3),
				unpack_index, numsek, numone, packed_count, unpack_count,
				(overhead + 7) >> 3, overhead);

		if (make_stat)
		{
			printf("Statistics%s\n\n", ena_rep_offset[compress_type] ? "               Offset: normal reused    =   all" : "");

			for (int what = 0; what < 3; what++)
			{
				int sum_own = 0;
				int sum_reu = 0;

				for (i = 1; i < 65536; i++)
					if (stat[what][0][i] || stat[what][1][i])
					{
						switch (what)
						{
						case UNPLEN: printf("  Unpack blocks  length:"); break;
						case SEKLEN: printf("  Sekvences with length:"); break;
						case SEKOFF: printf("  Sekvences with offset:"); break;
						}
						sum_own += stat[what][0][i];
						sum_reu += stat[what][1][i];
						printf("%-5u ... %5d", i, stat[what][0][i]);
						if (ena_rep_offset[compress_type]) printf(" %5d     = %5d", stat[what][1][i], stat[what][0][i] + stat[what][1][i]);
						putchar('\n');
					}
				if (what < 2)
				{
					int sum_ow2 = sum_own - stat[what][0][1];
					int sum_re2 = sum_reu - stat[what][1][1];
					printf("  ........................... ...\n");

					printf("  Summary for 2+ length only: ... %5d", sum_ow2);
					if (ena_rep_offset[compress_type]) printf(" %5d     = %5d", sum_re2, sum_ow2 + sum_re2);
					putchar('\n');

					printf("  Summary for 1+ length  all: ... %5d", sum_own);
					if (ena_rep_offset[compress_type]) printf(" %5d     = %5d", sum_reu, sum_own + sum_reu);
					putchar('\n');
				}
				putchar('\n');
			}

			for (int reused = 0; reused <= ena_rep_offset[compress_type]; reused++)
			{
				printf("Sekvence map");
				if (ena_rep_offset[compress_type]) printf(" with %s offset", reused ? "reused" : "normal");
				printf("\n\nOff\\Len =>");
				for (i = 1; limits[i] < 2 * unpack_index; i++) printf(" %5u", limits[i]); putchar('\n');
				for (offset = 1; limits[offset] < 2 * unpack_index; offset++)
				{
					printf("%5u ... ", limits[offset]);
					for (length = 1; limits[length] < 2 * unpack_index; length++)
						if (mapsek[offset][length][reused])
							printf(" %5u", mapsek[offset][length][reused]);
						else printf("     -");
					putchar('\n');
				}
				putchar('\n');
			}
		}

		if (make_unpack)
		{
			if (reverse_order)
				for (int i = (unpack_index >> 1) - 1; i >= 0; i--)
				{
					unsigned char tmp = unpack_data[i];
					unpack_data[i] = unpack_data[unpack_index - i - 1];
					unpack_data[unpack_index - i - 1] = tmp;
				}

			if (out_name) strcpy(unpack_name, out_name);
			if (extension)
			{
				int len = strlen(unpack_name);
				char * dotpos = strrchr(unpack_name + len - 5, '.');
				if (!dotpos)
				{
					unpack_name[len] = '.';
					dotpos = unpack_name + len;
				}
				strcpy(dotpos + 1, extension);
			}
			if (DBG) printf("Write %u bytes to file %s\n", unpack_index, unpack_name);

			FILE *unpack_file = fopen(unpack_name, "wb"); if (!unpack_file) { perror(unpack_name); continue; }
			if ((size_t)unpack_index != fwrite(unpack_data, 1, unpack_index, unpack_file)) printf("Can't write %s\n", unpack_name);
			fclose(unpack_file);
		}

		some_success++;
	}
	if (!number_of_files) puts("Error: No input files");
	else if (!make_stat && !make_list && !make_unpack && some_success) puts(
		"\n  You can use:\n"
		"   -d ... for depack\n"
		"   -l ... for listing\n"
		"   -s ... for statistics");

	return 0;
}
