///////////////////////////////////////////
// Busy soft // LzxPack 02 // 30.08.2021 //
///////////////////////////////////////////

// Boundary for all file adress dependent debug outputs
#define ENAZAC 0
#define ENAEND 65535

// Debug output level for used types of compressions: 0=none, 1=basic, 2=enhanced, 3=brutal...
#define ENAINF 0

// Debug output level for LZ compression engine: 0=none, 1=small, 2=large, 3=oops, 4=WTF?!, 5=... //
#define ENACOM 0

// Enable debug output of dump pack database
#define ENADMP 0

// Enable debug output of reversing sekvences
#define ENAREV 0

// Enable debug output of show pack model
#define ENADSP 0

// Enable debug output of generating output data
#define ENAOUT 0

// Enable misc statistics
#define ENASTA 0

// Enable more detailed time measure
#define ENATIM 0

// Debug level for optimizing pack model: 0=none, 1=summary, 2=brief, 2=detailed, 3=enhanced, 4=brutal, 5=...
//   OPTREP = optimize reused offsets
//   OPTUNP = optimize unpack blocks
#define OPTREP 0
#define OPTUNP 0

// Enable debug output of using sekvences for processing pack type
//   UNP = show initial unpack tax
//   ONE = show one byte sekvences
//   REP = repeated offset evaluation
//   CPT = capture sekvence
//   SUM = final summary
#define ENAUNP 0
#define ENAONE 0
#define ENAREP 0
#define ENACPT 0
#define ENASUM 0

// Select order number 0..N and name of compression for special displaying of sekvence selection (-1 = none)
#define SEKNUM -1
#define SEKNAM "LZX"

// 1=Force accepting all possible sekvences for testing purpuses, 2=Do not optimize sekvences
#define PCKALL 0

/////////////////////////////
// Compression definitions //
/////////////////////////////

#define COMCNT 10	// Number of COMxxx types

// Systems for coding unpacked data and sekvence lengths //

#define COMBLX 1	// Simple block compression - elias-gama length for both sekvence and unpacked block (real sequence 1+ length supports compress simple bytes)
#define COMBLC 2	// Simple block compression - elias-gama length for both sekvence and unpacked block (real sequence 2+ length has width more 1 bit: EG+1)
#define COMZX9 3	// Similar ZX7 compression - 1 bit for each unpacked byte and 1 bit + elias-gamma length for sekvence 1+ bytes (supports packing of simple bytes)
#define COMZX8 4	// Similar ZX7 compression - 1 bit for each unpacked byte and 1 bit + elias-gamma length for sekvence 2+ bytes (length EG+1 bit)
#define COMBS2 5	// Bitstream 1:unpack byte, 01:sek 2 bytes, 001:sek 3 bytes, 0001:sek 4+ bytes, 00000:1 unpack + rep offset, 00000:unpacked block 10+ bytes
#define COMBX1 6	// Bitstream 1:unpack, 01:sek2, 001:sek3, 0001:copied, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMBX2 7	// Bitstream 1:unpack, 01:sek2, 001:copied, 0001:sek3, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMBX3 8	// Bitstream 1:unpack, 01:copied, 001:sek2, 0001:sek3, 00001:sek4+, 000001:unpack + rep.offset, 0000001:copied + rep.offs, 0000000:block 12+
#define COMSX1 9	// Bitstream 1:sek 2+, 01:unpack, 001:copied, 0001:unpack + rep offset, 00001:copied + rep offset, 00000:unpack block 5+ bytes

// Systems for coding offsets in sequences //

#define POSOF1 1	// One fixed offset used for all sekvences
#define POSOF2 2	// Two fixed offsets (similar than ZX7)
#define POSOF3 3	// Three fixed step offsets (1,2,3 or 2,4,6 or 3,6,9 or 4,8,12 or 5,10,15)
#define POSOF4 4	// Four fixed step offsets (1,2,3,4 or 2,4,6,8 or 3,6,9,12 or 4,8,12,16 bits)
//efine POSxxx 5	// Not used (was OV0 but too complex and not used in comparation test)
#define POSOV1 6	// Simple variable elias-gama coding offset
#define POSOV2 7	// One fixed offset and one variable offset
#define POSOV3 8	// Two fixed offsets A,B and one variable elias-gama offset EG+B bits and +A addition
#define POSOV4 9	// Three fixed step offsets (1,2,3 or 2,4,6 or 3,6,9 or 4,8,12) and one variable offset

// Compression characteristics //

#define nie 0	// No
#define ANO 1	// Yes
// ...........................   -     1     2     3     4     5     6     7     8     9     -		// Compression identification
const char *compress_name[] = {":::","BLX","BLC","ZX9","ZX8","BS2","BX1","BX2","BX3","SX1","###"};	// Compression name identifier and abbreviation
const char use_bit_unpack[] = {  0,   nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,   0  };	// Compression uses one bit as mark for each unpacked byte
const char use_blk_unpack[] = {  0,   ANO,  ANO,  nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,   0  };	// Compression uses blocks with length for set of unpacked bytes
const char ena_first_byte[] = {  0,   nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,  ANO,   0  };	// Compressed data begins by one uncompressed byte without any overhead
const char ena_rep_offset[] = {  0,   nie,  nie,  nie,  nie,  ANO,  ANO,  ANO,  ANO,  ANO,   0  };	// Enable re-using the same offset value from previous sequence
const char max_1byte_sekv[] = {  0,     6,  nie,    6,  nie,  nie,    4,    5,    6,    6,   0  };	// Maximal bits in offset of one copied byte (0 = no packing single bytes)
const char endmark_length[] = { 127,   17,   17,   18,   17,   19,   20,   20,   20,   17,  127 };	// Number of overhead for end mark in compressed data

unsigned endmark_pattern[] = { (unsigned)-1,
	/* 1 BLX */ 0b00000000000000001,
	/* 2 BLC */ 0b00000000000000001,
	/* 3 ZX8 */ 0b00000000000000001,
	/* 4 ZX9 */ 0b00000000000000001,
	/* 5 BS2 */ 0b00000000000000001,
	/* 6 BX1 */ 0b01000000000000001,
	/* 7 BX2 */ 0b01000000000000001,
	/* 8 BX3 */ 0b01000000000000001,
	/* 9 SX1 */ 0b10000000000000001,
	(unsigned)-1 };

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define DELAY1 1000
#define DELAY2 250
#define MAXLEN 65536
#define DEPACKERS 128
#define PACKTYPES 3064

unsigned * TstAlloc(unsigned num);

unsigned idxout;
unsigned input_size;
unsigned bits_stream;
unsigned poke_stream;
unsigned packed_index;
unsigned packed_bytes;
unsigned unpack_bytes;
unsigned overhead_bits;
unsigned compressions;
unsigned depackers;

unsigned opti_com_cnt;
unsigned opti_pas_cnt;
unsigned opti_del_cnt;
unsigned opti_one_cnt;
unsigned opti_use_cnt;
unsigned opti_use_tax;
unsigned opti_rem_cnt;
unsigned opti_rem_tax;

unsigned compress_type;
unsigned compress_posi;
unsigned offset_copy;
unsigned offset1bits;
unsigned offset2bits;
unsigned reverse_order;
unsigned offset_limit;

char buffer[1024];
char nam_out[1024];
char ext_out[1024];

unsigned depacker_com[DEPACKERS];		// Depacker compress type
unsigned depacker_pos[DEPACKERS];		// Depacker offset coding
unsigned depacker_len[DEPACKERS][2];		// Depacker length: [0] for normal, [1] for reverse

unsigned char input_data[MAXLEN];		// Unpacked data from input file
unsigned char packed_data[MAXLEN << 1];	// Packed data to write into output file (packed data can be longer in bad case)

unsigned hash_bytes[0x00100];	// Hash table for quick searching of bytes for sekvences 1 byte long
unsigned hash_words[0x10000];	// Hash table for quick searching of words for sekvences 2+ bytes long

unsigned repper[MAXLEN];		// Period of repeated packed sequence ended in this byte in previous matches
unsigned replen[MAXLEN];		// Length of repeated packed sequence ended in this byte in previous matches
unsigned lstidx[MAXLEN];		// Locate of last processed sekvence matching with this position
unsigned lstlen[MAXLEN];		// Length of last processed sekvence matching with this position
unsigned bufnxt[MAXLEN];		// Lists of all found words - index of previous matching word
unsigned bytofs[MAXLEN];		// Offsets of bytes for evaluate and compress one copied bytes

unsigned typcom[PACKTYPES];		// Type of compression - coding length and unpacked data
unsigned typpos[PACKTYPES];		// Type of compression - coding offsets
unsigned declen[PACKTYPES];		// Depacked length (read from file in option -d)
unsigned maxlen[PACKTYPES];		// Maximal length of sekvence for compression
unsigned maxoff[PACKTYPES];		// Maximal offset of sekvence for compression
unsigned maxone[PACKTYPES];		// Maximal offset of single copied bytes
unsigned bytcop[PACKTYPES];		// Bit width of offset for one copied byte
unsigned ofset1[PACKTYPES];		// Bit width of offset 1 for sekvences (lower value)
unsigned ofset2[PACKTYPES];		// Bit width of offset 2 for sekvences (higher value)
unsigned locskp[PACKTYPES];		// Flag for skipping compression if best sekvence already found or offset is out of range
unsigned*limoff[PACKTYPES];		// Limits for offset
unsigned*taxofs[PACKTYPES];		// Taxes  for offset
unsigned*limlen[PACKTYPES];		// Limits for length
unsigned*taxlen[PACKTYPES];		// Taxes  for length
unsigned idxoff[PACKTYPES];		// Index into limoff/taxofs for actual position in input data
unsigned idxlen[PACKTYPES];		// Index into limlen/taxlen for actual position in input data
unsigned repkde[PACKTYPES];		// Index into data after sekvence with known offset .... used for quick find repated offset
unsigned repoff[PACKTYPES];		// Offset value of last known sekvence before repkde ... used for quick find repated offset
unsigned sumcom[PACKTYPES];		// Total length of packed data after LZ compression and adding end mark
unsigned sumall[PACKTYPES];		// Total length of packed data after optimalization
unsigned sumdec[PACKTYPES];		// Total length of packed data after optimalization and with depacker length
unsigned sumpck[PACKTYPES];		// Number of saved bytes - bytes copied by sekvences
unsigned sumunp[PACKTYPES];		// Number of unpacked bytes - bytes outside of sekvences
unsigned sumove[PACKTYPES];		// Overhead bits - total tax of all marks (unpack, sekvences, endmark)
unsigned sumsek[PACKTYPES];		// Number of sekvences in final pack model
unsigned mindst[PACKTYPES];		// Minimal distance between end of unpack and packed data during decompression

unsigned*limcom[COMCNT];		// Limits of length for compression type
unsigned*taxcom[COMCNT];		// Taxes  of length for compression type

unsigned unptax[COMCNT][MAXLEN];	// Taxes of unpack block lengths for used COMxxx compressions
unsigned lentax[COMCNT][MAXLEN];	// Taxes of   sekvence   lengths for used COMxxx compressions

unsigned ofstax[MAXLEN];	// Taxes   of sekvence offsets for actual processed POSxxx coding
uint16_t bufofs[MAXLEN];	// Offset  of packed sequecne leading and ended in this byte
uint16_t buflen[MAXLEN];	// Length  of packed sequecne leading and ended in this byte
uint32_t buftax[MAXLEN];	// Relative tax for unpacked optimization

uint16_t optzac[MAXLEN];	// List of possible optimizations for reused offset - first sekvence
uint16_t optend[MAXLEN];	// List of possible optimizations for reused offset -  last sekvence
uint16_t optben[MAXLEN];	// List of possible optimizations for reused offset - saved total tax

uint16_t memofs[MAXLEN][PACKTYPES];	// Offset  of packed sequecne ended in this byte
uint16_t memlen[MAXLEN][PACKTYPES];	// Length  of packed sequecne ended in this byte
uint32_t memtax[MAXLEN][PACKTYPES];	// Total price of packed data ended in this byte

#define relben repper	// Relative benefit of unpacked blocks for optimize unpacked

int GetNumberOffsets(int off)
{
	// Returns how many parameters is needed for selected offset coding //

	switch (off)
	{
	case POSOV1: return 0;
	case POSOF3:
	case POSOF4:
	case POSOV4:
	case POSOF1:
	case POSOV2: return 1;
	case POSOF2:
	case POSOV3: return 2;
	default: return -1;
	}
}

char * GetCompressType(char * txt, int comsel, int com = 15, int pos = 15, int one = 99, int of1 = 99, int of2 = 99)
{
	// Construct string with compression type -tnXYcNoAoB //

	if (comsel >= 0)
	{
		com = typcom[comsel];
		pos = typpos[comsel];
		one = bytcop[comsel];
		of1 = ofset1[comsel];
		of2 = ofset2[comsel];
	}
	txt += sprintf(txt, "-t%c%c%c", reverse_order ? 'r' : 'n', (com & 0x0F) | '0', (pos & 0x0F) | '0');
	if (!com || max_1byte_sekv[com]) txt += sprintf(txt, "c%u", one);
	int numofs = pos ? GetNumberOffsets(pos) : 2;
	if (numofs)
	{
		txt += sprintf(txt, "o%u", of1);
		if (numofs > 1)
			txt += sprintf(txt, "o%u", of2);
	}
	return txt;
}

void DisplayCompression(int comsel)
{
	// Show compression and offset coding type //

	int com = typcom[comsel];
	int pos = typpos[comsel];

	if ((unsigned)com > 9) printf("???");
	else printf(compress_name[com]);

	printf(" %c ", max_1byte_sekv[com] ? bytcop[comsel] | '0' : '-');
	int z = 0;

	switch (pos)
	{
	case POSOF1: z = printf("OF1 fixed offset %2u", ofset1[comsel]); break;
	case POSOF2: z = printf("OF2 offsets %2u %u", ofset1[comsel], ofset2[comsel]); break;
	case POSOF3: z = printf("OF3 offs %u %u %u", ofset1[comsel], 2 * ofset1[comsel], 3 * ofset1[comsel]); break;
	case POSOF4: z = printf("OF4 offs %u %u %u %u", ofset1[comsel], 2 * ofset1[comsel], 3 * ofset1[comsel], 4 * ofset1[comsel]); break;
	case POSOV1: z = printf("OV1 variable offset"); break;
	case POSOV2: z = printf("OV2 offsets %2u var", ofset1[comsel]); break;
	case POSOV3: z = printf("OV3 offs %2u %2u var", ofset1[comsel], ofset2[comsel]); break;
	case POSOV4: z = printf("OV4 offs %u %u %u var", ofset1[comsel], 2 * ofset1[comsel], 3 * ofset1[comsel]); break;
	default:	 z = printf("(unknown)");
	}
	printf(" %s ", "...................." + z);

	z = GetCompressType(buffer, comsel) - buffer;
	printf("%s %s ", buffer + 1, ".............." + z);
}

void SetOffsetTax(int comsel)
{
	// Set table of maximal values of sekvence offsets //
	//  and how many bits are needed for storing them  //

	unsigned o1 = ofset1[comsel];
	unsigned o2 = ofset2[comsel];

	unsigned *limit_offset = NULL;
	unsigned *taxes_offset = NULL;

	static const unsigned blk_off_lim[] = { 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned blk_off_tax[] = { 1, 3, 5,  7,  9, 11,  13,  15,  17,   19,   21,   23,   25,    27,    29,    31, 0 };

	switch (typpos[comsel])
	{
	case POSOV1:
		limit_offset = (unsigned*) blk_off_lim;
		taxes_offset = (unsigned*) blk_off_tax;
		break;

	case POSOF1:
		limit_offset = TstAlloc(4);
		limit_offset[0] = 1 << o1;
		limit_offset[1] = 0;
		taxes_offset = limit_offset + 2;
		taxes_offset[0] = o1;
		taxes_offset[1] = 0;
		break;

	case POSOF2:
		limit_offset = TstAlloc(6);
		limit_offset[0] = 1 << o1;
		limit_offset[1] = (1 << o1) + (1 << o2);
		limit_offset[2] = 0;
		taxes_offset = limit_offset + 3;
		taxes_offset[0] = o1 + 1;
		taxes_offset[1] = o2 + 1;
		taxes_offset[2] = 0;
		break;

	case POSOF3:
	{
		limit_offset = TstAlloc(8);
		limit_offset[0] = (1 << o1);
		limit_offset[1] = (1 << o1) + (1 << 2 * o1);
		limit_offset[2] = (1 << o1) + (1 << 2 * o1) + (1 << 3 * o1);
		limit_offset[3] = 0;
		taxes_offset = limit_offset + 4;
		taxes_offset[0] = 1 * o1 + 1;
		taxes_offset[1] = 2 * o1 + 2;
		taxes_offset[2] = 3 * o1 + 2;
		taxes_offset[3] = 0;
		break;
	}

	case POSOF4:
	{
		int offs = 0;
		limit_offset = TstAlloc(10); taxes_offset = limit_offset + 5;
		for (int i = 0; i < 4; i++)
		{
			int bits = ((i + 1) * o1);
			offs += 1 << bits;
			limit_offset[i] = offs;
			taxes_offset[i] = bits + 2;
		}
		limit_offset[4] = 0;
		taxes_offset[4] = 0;
		break;
	}

	case POSOV4:
	{
		int i = 0;
		int s = 0;
		int bits = 0;
		int offs = 0;
		limit_offset = TstAlloc(24 - 3 * o1);
		taxes_offset = TstAlloc(24 - 3 * o1);
		for (i = 0; i < 3; i++)
		{
			bits = ((i + 1) * o1);
			offs += 1 << bits;
			limit_offset[s] = offs;
			taxes_offset[s] = bits + 2;
			s++;
		}
		int limit = offs;
		i = 0;
		do
		{
			limit += 1 << (i + bits);
			limit_offset[s] = limit;
			taxes_offset[s] = bits + 2 * i + 3;
			s++;
			i++;
		} while (limit < 65536);
		limit_offset[s] = 0;
		taxes_offset[s] = 0;
		break;
	}

	case POSOV2:
	{
		limit_offset = TstAlloc(20 - o1);
		taxes_offset = TstAlloc(20 - o1);
		limit_offset[0] = 1 << o1;
		taxes_offset[0] = o1 + 1;
		int limit = 0;
		int i = 1;
		do
		{
			limit = (1 << (i + o1));
			limit_offset[i] = limit;
			taxes_offset[i] = 2 * i + o1;
			i++;
		} while (limit < 65536);
		limit_offset[i] = 0;
		taxes_offset[i] = 0;
		break;
	}

	case POSOV3:
	{
		limit_offset = TstAlloc(20 - o2);
		limit_offset[0] = 1 << o1;
		limit_offset[1] = (1 << o1) + (1 << o2);
		taxes_offset = TstAlloc(20 - o2);
		taxes_offset[0] = o1 + 1;
		taxes_offset[1] = o2 + 2;
		int limit = 0;
		int i = 0;
		do
		{
			limit = (1 << o1) + (1 << (i + o2 + 1));
			limit_offset[i + 2] = limit;
			taxes_offset[i + 2] = 2 * i + 3 + o2;
			i++;
		} while (limit < 65536);
		limit_offset[i + 2] = 0;
		taxes_offset[i + 2] = 0;
		break;
	}
	}

	limoff[comsel] = limit_offset;
	taxofs[comsel] = taxes_offset;
}

void SetSekvenceLengthTax(unsigned comtyp)
{
	// Set table of maximal values of sekvence lengths //
	//  and how many bits are needed for storing them  //

	unsigned * limit_length = NULL;
	unsigned * taxes_length = NULL;

	static const unsigned blc_len_lim[] = { 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned blc_len_tax[] = { 3, 5,  7,  9, 11,  13,  15,  17,   19,   21,   23,   25,    27,    29,    31, 0 };

	static const unsigned zxs_len_lim[] = { 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned zxs_len_tax[] = { 2, 4, 6,  8, 10, 12,  14,  16,  18,   20,   22,   24,   26,    28,    30,    32, 0 };

	static const unsigned bs2_len_lim[] = { 2, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned bs2_len_tax[] = { 2, 3, 7,  9, 11, 13,  15,  17,  19,   21,   23,   25,   27,    29,    31,    33, 0 };

	static const unsigned bx1_len_lim[] = { 1, 2, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned bx1_len_tax[] = { 4, 2, 3, 8, 10, 12, 14,  16,  18,  20,   22,   24,   26,   28,    30,    32,    34, 0 };

	static const unsigned bx2_len_lim[] = { 1, 2, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned bx2_len_tax[] = { 3, 2, 4, 8, 10, 12, 14,  16,  18,  20,   22,   24,   26,   28,    30,    32,    34, 0 };

	static const unsigned bx3_len_lim[] = { 1, 2, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned bx3_len_tax[] = { 2, 3, 4, 8, 10, 12, 14,  16,  18,  20,   22,   24,   26,   28,    30,    32,    34, 0 };

	static const unsigned sx1_len_lim[] = { 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static const unsigned sx1_len_tax[] = { 3, 3, 5,  7,  9, 11, 13,   15,  17,   19,   21,   23,   25,    27,    29,    31, 0 };

	switch (comtyp)
	{
	case COMBLX: limit_length = (unsigned*)zxs_len_lim; taxes_length = (unsigned*)zxs_len_tax; break;
	case COMBLC: limit_length = (unsigned*)blc_len_lim; taxes_length = (unsigned*)blc_len_tax; break;
	case COMZX9: limit_length = (unsigned*)zxs_len_lim; taxes_length = (unsigned*)zxs_len_tax; break;
	case COMZX8: limit_length = (unsigned*)blc_len_lim; taxes_length = (unsigned*)blc_len_tax; break;
	case COMBS2: limit_length = (unsigned*)bs2_len_lim; taxes_length = (unsigned*)bs2_len_tax; break;
	case COMBX1: limit_length = (unsigned*)bx1_len_lim; taxes_length = (unsigned*)bx1_len_tax; break;
	case COMBX2: limit_length = (unsigned*)bx2_len_lim; taxes_length = (unsigned*)bx2_len_tax; break;
	case COMBX3: limit_length = (unsigned*)bx3_len_lim; taxes_length = (unsigned*)bx3_len_tax; break;
	case COMSX1: limit_length = (unsigned*)sx1_len_lim; taxes_length = (unsigned*)sx1_len_tax; break;
	}

	limcom[comtyp] = limit_length;
	taxcom[comtyp] = taxes_length;

	unsigned limzac;
	unsigned * comtax = lentax[comtyp];

	comtax[0] = 880000000;
	comtax[1] = 770000000;
	limzac = *limit_length;
	if (limzac > 2) limzac = 2;
	while (*limit_length)
	{
		unsigned limend = *limit_length;
		unsigned limtax = *taxes_length;
		for (unsigned i = limzac; i <= limend; i++) comtax[i] = limtax;
		limzac = limend + 1;
		limit_length++;
		taxes_length++;
	}
	comtax[limzac] = 990000000;
}

void SetUnpackLengthTax(unsigned comtyp)
{
	// Set arrays for total tax (data + overhead) of not compressed data block //

	unsigned * comtax = unptax[comtyp];
	unsigned i, lim;
	unsigned idx = 1;
	unsigned mul = 9;
	unsigned lin = 0;
	unsigned blk = 0;
	unsigned tax = 0;

	switch (comtyp)
	{
	case COMBLX:
	case COMBLC:
		lin = 0;
		blk = 1;
		tax = 2;
		break;

	case COMZX8:
	case COMZX9:
		lin = input_size;
		break;

	case COMBS2:
		lin = 9;
		blk = 8;
		tax = 9;
		break;

	case COMBX1:
	case COMBX2:
	case COMBX3:
		lin = 11;
		blk = 8;
		tax = 11;
		break;

	case COMSX1:
		mul = 10;
		lin = 4;
		blk = 4;
		tax = 8;
		break;

	default: return;
	}

	*comtax = 0;
	for (idx = 1; idx <= lin; idx++) comtax[idx] = mul * idx;
	while (idx <= input_size)
	{
		lim = blk;
		if (idx + lim > input_size) lim = input_size - idx + 1;
		for (i = 0; i < lim; i++)
		{
			comtax[idx] = tax + (idx << 3);
			idx++;
		}
		blk <<= 1;
		tax += 2;
	}
	if (idx < 65536) comtax[idx] = 990000000;
}

unsigned GetSekvenceTax(unsigned comsel, unsigned offset, unsigned length)
{
	// Get tax or overhead of sekvence with normal own offset //

	if (length < 1 || offset < 1 || offset == 0xFFFF)
	{
		printf("Error: GetSekvenceTax:  Offset:%u  Length:%u  ", offset, length);
		if (compressions > 1) DisplayCompression(comsel);
		putchar('\n');
		exit(1);
	}

	if (length == 1 && !max_1byte_sekv[typcom[comsel]])
	{
		printf("Error: GetSekvenceTax:  Not supported one copied bytes in ");
		if (compressions > 1) DisplayCompression(comsel);
		putchar('\n');
		exit(1);
	}

	if (length == 1) return taxlen[comsel][0] + bytcop[comsel];

	unsigned o, *offsets = limoff[comsel]; for (o = 0; offsets[o]; o++) if (offset <= offsets[o]) break;

	return taxofs[comsel][o] + lentax[typcom[comsel]][length];
}

int GetBitsNumber(int number)
{
	// Get bit width of number //

	if (number < 1) return 0;
	if (number < 2) return 1;
	if (number < 4) return 2;
	if (number < 8) return 3;
	if (number < 16) return 4;
	if (number < 32) return 5;
	if (number < 64) return 6;
	if (number < 128) return 7;
	if (number < 256) return 8;
	if (number < 512) return 9;
	if (number < 1024) return 10;
	if (number < 2048) return 11;
	if (number < 4096) return 12;
	if (number < 8192) return 13;
	if (number < 16384) return 14;
	if (number < 32768) return 15;
	if (number < 65536) return 16;
	return 999999;
}

unsigned GetUnpReusedOffsetSekTax(unsigned comsel, unsigned lenunp, unsigned byte_offset, unsigned lensek)
{
	// Get all tax of unpacked data block or one copied byte + sekvence with reused offset //

	if (lenunp < 1 || lensek < 2 || byte_offset == 0xFFFF)
	{
		printf("Error: GetUnpReusedOffsetSekTax:  Bad parameters LenUnp:%d  BytOff:%d  LenSek:%d   ", lenunp, byte_offset, lensek);
		if (compressions > 1) DisplayCompression(comsel);
		putchar('\n');
		return 990000;
	}

	switch (typcom[comsel])
	{
	case COMBS2:
		if (lenunp <= 10)
			return 9 * (lenunp - 1) + 5 + 8 + 2 * GetBitsNumber(lensek) - 2;
		else
			return 2 * GetBitsNumber(lenunp - 2 - 1) + 1 + 8 * (lenunp - 1) + 5 + 8 + 2 * GetBitsNumber(lensek) - 2;

	case COMBX1:
	case COMBX2:
	case COMBX3:
		if (byte_offset)
			return 7 + bytcop[comsel] + 2 * GetBitsNumber(lensek) - 2;
		else if (lenunp <= 12)
			return 9 * (lenunp - 1) + 6 + 8 + 2 * GetBitsNumber(lensek) - 2; // unp=2 sek=2 ... 10+4+8+2
		else
			return 2 * GetBitsNumber(lenunp - 5) + 3 + 8 * (lenunp - 1) + 6 + 8 + 2 * GetBitsNumber(lensek) - 2;

	case COMSX1:
		if (byte_offset)
			return 5 + bytcop[comsel] + 2 * GetBitsNumber(lensek) - 2;
		else if (lenunp <= 5)
			return 10 * (lenunp - 1) + 4 + 8 + 2 * GetBitsNumber(lensek) - 2; // unp=2 sek=2 ... 10+4+8+2
		else
			return 2 * GetBitsNumber(lenunp - 2) + 2 + 8 * (lenunp - 1) + 4 + 8 + 2 * GetBitsNumber(lensek) - 2;
	}
	printf("Error: GetUnpReusedOffsetSekTax:  Bad compress type %u in compression %u\n", typcom[comsel], comsel);
	exit(13);
}

void PutBit(unsigned bit)
{
	// Store one bit into bitstream in packed data //

	if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) printf(" %c", bit ? '1':'0');
	if (!bits_stream)
	{
		poke_stream = packed_index;
		packed_index++;
		bits_stream++;
	}
	bits_stream = (bits_stream << 1) | (bit ? 1 : 0);
	if (bits_stream & 0xFF00)
	{
		packed_data[poke_stream] = (unsigned char)(bits_stream & 0xFF);
		bits_stream = 0;
	}
	overhead_bits++;
}

void PutValue(unsigned value, unsigned bits)
{
	// Store "bits" bits of binary value "value" //

	if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) printf(" (Val:%u B:%u)", value, bits);
	if (value >= (1u << bits)) { printf("\nError in PutValue: Input value %u overflows %u for %u bits.\n", value, (1 << bits) - 1, bits); exit(7); }
	if (bits) for (unsigned mask = 1u << (bits - 1); mask; mask >>= 1) PutBit(value & mask);
}

void PutEliasGama(unsigned value, unsigned bits)
{
	// Store Elias-gama coded value //
	// <n*zero> 1 <n+bits of value> //

	if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) printf(" (Eli:%u B:%u)", value, bits);
	if (value < (1u << bits)) { printf("\nError in PutEliasGama: Input value %u underflows %u for %u bits.\n", value, 1 << bits, bits); exit(7); }
	while (value > (1u << (bits + 1)) - 1) { bits++; PutBit(0); }
	PutValue(value, bits + 1);
}

void PutSekvence(unsigned comsel, unsigned offset, unsigned length)
{
	// Store sekvence with own offset into output packed data //

	if ((length > 1 && (offset > maxoff[comsel] || length > maxlen[comsel]))
		|| (length == 1 && (!max_1byte_sekv[typcom[comsel]] || offset > (1u << bytcop[comsel])))
		|| length < 1
		|| offset == 0xFFFF)
	{
		printf("\nError in PutSekvence: Offset:%u  Length:%u  MaxOff:%u  MaxOne:%u  MaxLen:%u in ",
			offset, length, maxoff[comsel], 1 << bytcop[comsel], maxlen[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(6);
	}

	if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) printf(" (sek)");

	switch (typcom[comsel])
	{
	case COMZX8:
	case COMBLC:
		PutBit(0);
		PutEliasGama(length, 1);
		break;

	case COMBLX:
		PutEliasGama(length, 0);
		PutBit(0);
		break;

	case COMZX9:
		PutBit(0);
		PutEliasGama(length, 0);
		break;

	case COMBS2:
		if (length == 2)
		{
			PutBit(0);
			PutBit(1);
		}
		else if (length == 3)
		{
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutEliasGama(length, 2);
		}
		break;

	case COMBX1:
		if (length == 1)
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else if (length == 2)
		{
			PutBit(0);
			PutBit(1);
		}
		else if (length == 3)
		{
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutEliasGama(length, 2);
		}
		break;

	case COMBX2:
		if (length == 1)
		{
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else if (length == 2)
		{
			PutBit(0);
			PutBit(1);
		}
		else if (length == 3)
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutEliasGama(length, 2);
		}
		break;

	case COMBX3:
		if (length == 1)
		{
			PutBit(0);
			PutBit(1);
		}
		else if (length == 2)
		{
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else if (length == 3)
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutEliasGama(length, 2);
		}
		break;

	case COMSX1:
		if (length == 1)
		{
			PutBit(0);
			PutBit(0);
			PutBit(1);
		}
		else
		{
			PutBit(1);
			PutEliasGama(length, 1);
		}
		break;

	default:
		printf("\nError in PutSekvence: Wrong compress %u: ", typcom[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(12);
	}

	if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) printf(" (ofs)");

	if (length == 1)
	{
		if (offset > (1u << bytcop[comsel]))
		{
			printf("\nError in PutSekvence: Offset %u for one byte overflows max value %u in ", offset, 1u << bytcop[comsel]);
			DisplayCompression(comsel);
			putchar('\n');
			exit(6);
		}
		PutValue(offset - 1, bytcop[comsel]);
		return;
	}

	if (offset > maxoff[comsel])
	{
		printf("\nError in PutSekvence: Offset %u for sekvence overflows max value %u in ", offset, maxoff[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(6);
	}

	unsigned i;
	unsigned o1 = ofset1[comsel];
	unsigned o2 = ofset2[comsel];

	switch (typpos[comsel])
	{
	case POSOV1:
		PutEliasGama(offset, 0);
		break;

	case POSOF1:
		PutValue(offset - 1, o1);
		break;

	case POSOF2:
		if (offset <= (1u << o1))
		{
			PutBit(1);
			PutValue(offset - 1, o1);
		}
		else
		{
			PutBit(0);
			PutValue(offset - (1u << o1) - 1, o2);
		}
		break;

	case POSOV2:
		if (offset <= (1u << o1))
		{
			PutBit(0);
			PutValue(offset - 1, o1);
		}
		else
		{
			PutBit(1);
			PutEliasGama(offset - 1, o1);
		}
		break;

	case POSOV3:
		if (offset <= (1u << o1))
		{
			PutBit(0);
			PutValue(offset - 1, o1);
		}
		else if (offset <= (1u << o1) + (1 << o2))
		{
			PutBit(1);
			PutBit(0);
			PutValue(offset - (1u << o1) - 1, o2);
		}
		else
		{
			PutBit(1);
			PutBit(1);
			PutEliasGama(offset - (1u << o1) - 1, o2);
		}
		break;

	case POSOF3:
		for (i = 0; i < 2; i++) if (offset <= limoff[comsel][i] || !limoff[comsel][i]) break;
		if (i) offset -= limoff[comsel][i - 1];
		if (!i)
			PutBit(0);
		else
		{
			PutBit(1);
			PutBit(!(i & 0x01));
		}
		PutValue(offset - 1, o1*(i + 1));
		break;

	case POSOV4:
		for (i = 0; i < 3; i++) if (offset <= limoff[comsel][i] || !limoff[comsel][i]) break;
		if (i && i < 3) offset -= limoff[comsel][i - 1];
		PutBit(i & 0x02);
		PutBit(i & 0x01);
		if (i < 3) PutValue(offset - 1, o1*(i + 1));
		else PutEliasGama(offset - 1 - limoff[comsel][i - 2], 3 * o1);
		break;

	case POSOF4:
		for (i = 0; i < 4; i++) if (offset <= limoff[comsel][i] || !limoff[comsel][i]) break;
		if (i) offset -= limoff[comsel][i - 1];
		PutBit(i & 0x02);
		PutBit(i & 0x01);
		PutValue(offset - 1, o1*(i + 1));
		break;

	default:
		printf("\nError in PutSekvenceOffset: Wrong offset type %u for ", typpos[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(12);
	}
}

void PutUnpackData(unsigned comsel, unsigned kdesme, unsigned length)
{
	// Store block of unpacked data into output packed data //

	if (length < 1)
	{
		printf("\nError in PutUnpackData: Wrong length %d in ", length);
		DisplayCompression(comsel);
		putchar('\n');
		exit(6);
	}

	if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (nop)");

	switch (typcom[comsel])
	{
	case COMBLX:
		PutEliasGama(length, 0);
		PutBit(1);
		memcpy(packed_data + packed_index, input_data + kdesme, length);
		packed_index += length;
		break;

	case COMBLC:
		PutBit(1);
		PutEliasGama(length, 0);
		memcpy(packed_data + packed_index, input_data + kdesme, length);
		packed_index += length;
		break;

	case COMZX8:
	case COMZX9:
		for (unsigned i = 0; i < length; i++)
		{
			PutBit(1);
			packed_data[packed_index++] = input_data[kdesme + i];
		}
		break;

	case COMBS2:
		if (length < 10)
		{
			for (unsigned i = 0; i < length; i++)
			{
				PutBit(1);
				packed_data[packed_index++] = input_data[kdesme + i];
			}
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutEliasGama(length - 2, 3);
			memcpy(packed_data + packed_index, input_data + kdesme, length);
			packed_index += length;
		}
		break;

	case COMBX1:
	case COMBX2:
	case COMBX3:
		if (length < 12)
		{
			for (unsigned i = 0; i < length; i++)
			{
				PutBit(1);
				packed_data[packed_index++] = input_data[kdesme + i];
			}
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutEliasGama(length - 4, 3);
			memcpy(packed_data + packed_index, input_data + kdesme, length);
			packed_index += length;
		}
		break;

	case COMSX1:
		if (length < 5)
		{
			for (unsigned i = 0; i < length; i++)
			{
				PutBit(0);
				PutBit(1);
				packed_data[packed_index++] = input_data[kdesme + i];
			}
		}
		else
		{
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutEliasGama(length - 1, 2);
			memcpy(packed_data + packed_index, input_data + kdesme, length);
			packed_index += length;
		}
		break;

	default:
		printf("\nError in PutUnpackData: Wrong compress %u: ", typcom[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(13);
	}
}

void PutReusedOffset(unsigned comsel, unsigned kdesme, unsigned unpack_length, unsigned byte_offset, unsigned packed_length)
{
	// Store block of unpacked data or copied byte, and then sekvence with reused offset into output packed data //

	if (!unpack_length || packed_length < 2 || (byte_offset && unpack_length != 1))
	{
		printf("\nError in PutReusedOffset at +%u %04X for  ", kdesme, kdesme);
		DisplayCompression(comsel);
		printf("\n"
			"  Unpack length: %d\n"
			"  Copied offset: %d\n"
			"  Packed length: %d\n",
			unpack_length, byte_offset, packed_length);
		exit(6);
	}

	switch (typcom[comsel])
	{
	case COMBS2:
		if (unpack_length > 1) PutUnpackData(comsel, kdesme, unpack_length - 1);
		if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs:unp)");
		PutBit(0);
		PutBit(0);
		PutBit(0);
		PutBit(0);
		PutBit(1);
		packed_data[packed_index] = input_data[kdesme + unpack_length - 1];
		if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("  [data:%02X] ", packed_data[packed_index]);
		packed_index++;
		PutEliasGama(packed_length, 1);
		break;

	case COMBX1:
	case COMBX2:
	case COMBX3:
		if (unpack_length > 1) PutUnpackData(comsel, kdesme, unpack_length - 1);
		if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs)");
		if (byte_offset)
		{
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs:cpy)");
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutValue(byte_offset - 1, bytcop[comsel]);
		}
		else
		{
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs:unp)");
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (unp)");
			packed_data[packed_index] = input_data[kdesme + unpack_length - 1];
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("  [data:%02X]", packed_data[packed_index]);
			packed_index++;
			kdesme++;
		}
		PutEliasGama(packed_length, 1);
		break;

	case COMSX1:
		if (unpack_length > 1) PutUnpackData(comsel, kdesme, unpack_length - 1);
		if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs)");
		if (byte_offset)
		{
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs:cpy)");
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			PutValue(byte_offset - 1, bytcop[comsel]);
		}
		else
		{
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (repofs:unp)");
			PutBit(0);
			PutBit(0);
			PutBit(0);
			PutBit(1);
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" (unp)");
			packed_data[packed_index] = input_data[kdesme + unpack_length - 1];
			if (ENAOUT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("  [data:%02X] ", packed_data[packed_index]);
			packed_index++;
		}
		PutEliasGama(packed_length, 1);
		break;

	default:
		printf("\nError in PutReusedOffset: Wrong compress %u: ", typcom[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(12);
	}
}
void SetCompression(unsigned compress_type, unsigned compress_offset, unsigned offone, unsigned off1, unsigned off2)
{
	// Creates all needed tables for one type of compression //

	unsigned i;
	if (compressions >= PACKTYPES) {compressions++; return; }
	if (compress_type < 1 || compress_type > 9 || compress_offset < 1 || compress_offset > 9) return;

	typcom[compressions] = compress_type;
	typpos[compressions] = compress_offset;
	bytcop[compressions] = offone;
	ofset1[compressions] = off1;
	ofset2[compressions] = off2;

	SetOffsetTax(compressions);

	limlen[compressions] = limcom[compress_type];
	taxlen[compressions] = taxcom[compress_type];
	for (i = 1; limoff[compressions][i]; i++); maxoff[compressions] = limoff[compressions][i - 1];
	for (i = 1; limlen[compressions][i]; i++); maxlen[compressions] = limlen[compressions][i - 1];

	maxone[compressions] = 1 << offone;

	declen[compressions] = 0;
	if (depackers)
	{
		for (unsigned i = 0; i < depackers; i++)
			if (compress_type   == depacker_com[i]
			 && compress_offset == depacker_pos[i])
			{
				declen[compressions] = depacker_len[i][reverse_order];
				break;
			}
	}

	if (ENAINF)
	{
		printf("%5u ... ", compressions);
		DisplayCompression(compressions);
		printf("MaxLen:%-5u  MaxSekOfs:%-5u  MaxOneOfs:%u  RepOfs:%s",
			maxlen[compressions],
			maxoff[compressions],
			max_1byte_sekv[typcom[compressions]],
			ena_rep_offset[compress_type] ? "YES":"no ");

		if (declen[compressions]) printf("    DepLen:%u", declen[compressions]);
		putchar('\n');
		if (ENAINF > 1)
		{
			printf("\t\tOffset limits: "); for (int i = 0; limoff[compressions][i]; i++) printf("  %u.%u", limoff[compressions][i], taxofs[compressions][i]); putchar('\n');
			printf("\t\tLength limits: "); for (int i = 0; limlen[compressions][i]; i++) printf("  %u.%u", limlen[compressions][i], taxlen[compressions][i]); putchar('\n');
			putchar('\n');
		}
	}
	compressions++;
}

void PrepareAllCompressionTypes(unsigned input_size)
{
	// Generates set of compressions according to specified compression type by input cmdline parameter -tXY... //

	unsigned maxbtin = GetBitsNumber(input_size - 1);	// Maximal bits of offsets due to length of input file
	unsigned maxbtof = GetBitsNumber(offset_limit - 1);	// Maximal bits of offsets due to selected maximal dictionary size

	if (!maxbtof) maxbtof = 1;
	unsigned maxbits = maxbtin > maxbtof ? maxbtof : maxbtin; // Maximal bits of offsets used for generating compressions

	unsigned minstep = 1;	// Minimal bits of step offsets in OF3 OF4 OV4
	unsigned mincopy = 1;	// Minimal bits of offset for copying one byte

	unsigned of2_o1min = 1;				// Minimal bits of smaler offset in OF2
	unsigned of2_o2min = 3;				// Minimal bits of bigger offset in OF2
	unsigned of2_dista = 2;				// Minimal bits of distance between smaller and bigger offset for OF2
	unsigned of2_o1max = maxbits - 2;	// Maximal bits of smaler offset in OF2
	unsigned ov3_o1min = 1;				// Minimal bits of smaler offset in OV3
	unsigned ov3_o2min = 3;				// Minimal bits of bigger offset in OV3
	unsigned ov3_o1max = maxbits - 4;	// Maximal bits of smaler offset in OV3

	// Reduction range of parameters to keep usable memory requirements //
	// On 64bit environment this is not needed. In case of compile for  //
	// 64 bit, remove this following IF and set PACKTYPES to value 8286 //
	
	if (!compress_type && !offset_copy && !offset1bits && !offset2bits)
	{
		if (compress_posi == POSOF2)
		{
			switch (maxbits)
			{
			case 16: of2_o1max = 10; of2_o2min = 7; break;
			case 15: of2_o1max = 11; of2_o2min = 5; break;
			}
		}
		if (!compress_posi)
		{
			switch (maxbits)
			{
			case 16: of2_o1min = 7; of2_o1max = 12; of2_o2min = 10; of2_dista = 5; ov3_o1min = 5; ov3_o1max = 11; ov3_o2min = 8; minstep = 2; mincopy = 3; break; // 65535 //
			case 15: of2_o1min = 6; of2_o1max = 11; of2_o2min = 9;  of2_dista = 5; ov3_o1min = 4; ov3_o1max = 10; ov3_o2min = 7; minstep = 2; mincopy = 3; break; // 32768 //
			case 14: of2_o1min = 5; of2_o1max = 10; of2_o2min = 8;  of2_dista = 4; ov3_o1min = 3; ov3_o1max = 9;  ov3_o2min = 6; minstep = 2; mincopy = 3; break; // 16384 //
			case 13: of2_o1min = 4; of2_o1max = 9;  of2_o2min = 7;  of2_dista = 3; ov3_o1min = 2; ov3_o1max = 9;  ov3_o2min = 5; minstep = 2; mincopy = 3; break; //  8192 //
			case 12: of2_o1min = 3; of2_o1max = 8;  of2_o2min = 7;  of2_dista = 3; ov3_o1min = 2; ov3_o1max = 8;  ov3_o2min = 4; minstep = 1; mincopy = 2; break; //  4096 //
			case 11: of2_o1min = 2; of2_o1max = 7;  of2_o2min = 6;  of2_dista = 2; ov3_o1min = 1; ov3_o1max = 7;  ov3_o2min = 3; minstep = 1; mincopy = 2; break; //  2048 //
			}
		}
	} // End of IF needed for 32 bit environment //
		
	if (ENAINF)
		printf(
			"\nUsed compression types for -tx%c%cc%uo%uo%u   FileLen: %u (%u)  OfsLim: %u (%u)  =>  "
			"OF2  Ofs1: %u-%u  Ofs2: %u-%u  Dif:%u    "
			"OV3  Ofs1: %u-%u  Ofs2: %u-%u    "
			"MinStep: %u  MinCopy: %u\n\n",
			compress_type + '0', compress_posi + '0', offset_copy, offset1bits, offset2bits,
			input_size, maxbtin, offset_limit, maxbtof,
			of2_o1min, of2_o1max, of2_o2min, maxbits, of2_dista,
			ov3_o1min, ov3_o1max, ov3_o2min, maxbits-2,
			minstep, mincopy);

	for (unsigned com = compress_type ? compress_type : 1; com <= (compress_type ? compress_type : COMCNT - 1); com++)
	{
		SetUnpackLengthTax(com);	// Create arrays of total tax of unpack block //
		SetSekvenceLengthTax(com);	// Create arrays of total tax of sekvence length //
	}

	if (ENAINF > 1)
		for (int what = 0; what < 2; what++) // 0=unpack, 1=sekvence //
		{
			printf("%s overhead   (bits:min-max)\n\n", what ? "SekLen tax" : "Unpack tax");
			for (unsigned com = compress_type ? compress_type : 1; com <= (compress_type ? compress_type : COMCNT - 1); com++)
			{
				printf("  %u %s  ...", com, compress_name[com]);
				unsigned max = 24;
				unsigned idx = (what && !max_1byte_sekv[com]) ? 2 : 1;
				unsigned * comtax = what ? lentax[com] : unptax[com];

				while (max && idx <= input_size)
				{
					unsigned nxt = idx;
					unsigned bits = comtax[nxt] - (what ? 0 : (nxt << 3));
					while (nxt <= input_size && bits == comtax[nxt] - (what ? 0 : (nxt << 3))) nxt++;
					printf("  %d:%u-%u", bits, idx, nxt - 1);
					idx = nxt;
					max--;
				}
				if (!max && idx <= input_size) printf("  ...");
				putchar('\n');
			}
			putchar('\n');
		}
	
	compressions = 0;

	for (unsigned com = compress_type ? compress_type : 1; com <= (compress_type ? compress_type : COMCNT - 1); com++)
	{
		// Bit limits for compressing one byte //
		unsigned max_1sek = max_1byte_sekv[com];
		unsigned min_cof2 = max_1sek ? (offset_copy ? offset_copy : mincopy) : 0;
		unsigned min_copy = max_1sek ? (offset_copy ? offset_copy : 1) : 0;
		unsigned max_copy = max_1sek ? (offset_copy ? offset_copy : (max_1sek > maxbits ? maxbits : max_1sek)) : 0;

		if (ENAINF > 1) printf("OffsetCopy:%u  ComPress:%u  MinCopy:%u:%u  MaxCopy:%u:%u:%u\n\n", offset_copy, com, min_copy, min_cof2, max_1sek, maxbits, max_copy);

		if ((compress_type && compress_posi) || (com > 0 && com < COMCNT))
		{
			if (!compress_posi || compress_posi == POSOF1)
				for (unsigned i = offset1bits ? offset1bits : 1; i <= (offset1bits ? offset1bits : maxbits); i++)
					for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
						SetCompression(com, POSOF1, byte_offset, i, 0);

			if (!compress_posi || compress_posi == POSOF2)
			{
				if (offset1bits) of2_o2min = offset1bits + 2;
				for (unsigned i = offset2bits ? offset2bits : of2_o2min; i <= (offset2bits ? offset2bits : maxbits); i++)
				{
					int lim = of2_o1max; if (lim > (int)(i - of2_dista)) lim = i - of2_dista;
					for (unsigned j = offset1bits ? offset1bits : of2_o1min; (int)j <= (int)(offset1bits ? offset1bits : lim); j++)
						for (unsigned byte_offset = min_cof2; byte_offset <= max_copy; byte_offset++)
							SetCompression(com, POSOF2, byte_offset, j, i);
				}
			}

			if (!compress_posi || compress_posi == POSOF3)
			{
				if (offset1bits > 5 && compress_posi) { printf("Error: Offset %u overflows maximal value 5 for OF3\n", offset1bits); exit(1); }
				if (offset1bits <= 5)
					for (unsigned i = offset1bits ? offset1bits : minstep; i <= (offset1bits ? offset1bits : (maxbits + 1u) / 3); i++)
						for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
							SetCompression(com, POSOF3, byte_offset, i, 0);
			}

			if (!compress_posi || compress_posi == POSOF4)
			{
				if (offset1bits > 4 && compress_posi) { printf("Error: Offset %u overflows maximal value 4 for OF4\n", offset1bits); exit(1); }
				if (offset1bits <= 4)
					for (unsigned i = offset1bits ? offset1bits : minstep; i <= (offset1bits ? offset1bits : (maxbits + 1u) >> 2); i++)
						for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
							SetCompression(com, POSOF4, byte_offset, i, 0);
			}

			if (!compress_posi || compress_posi == POSOV1)
				for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
					SetCompression(com, POSOV1, byte_offset, 0, 0);

			if (!compress_posi || compress_posi == POSOV2)
				for (unsigned i = offset1bits ? offset1bits : 1; (int)i <= (offset1bits ? (int)offset1bits : (int)maxbits - 2); i++)
					for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
						SetCompression(com, POSOV2, byte_offset, i, 0);

			if (!compress_posi || compress_posi == POSOV3)
			{
				if (offset1bits) ov3_o2min = offset1bits + 2;
				for (unsigned i = offset2bits ? offset2bits : ov3_o2min; (int)i <= (offset2bits ? (int)offset2bits : (int)maxbits - (int)2); i++)
				{
					int lim = ov3_o1max; if (lim > (int)i - 2) lim = i - 2;
					for (unsigned j = offset1bits ? offset1bits : ov3_o1min; (int)j <= (int)(offset1bits ? offset1bits : lim); j++)
						for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
							SetCompression(com, POSOV3, byte_offset, j, i);
				}
			}

			if (!compress_posi || compress_posi == POSOV4)
			{
				if (offset1bits > 4 && compress_posi) { printf("Error: Offset %u overflows maximal value 4 for OV4\n", offset1bits); exit(1); }
				if (offset1bits <= 4)
					for (unsigned i = offset1bits ? offset1bits : minstep; i <= (offset1bits ? offset1bits : (maxbits + 1u) >> 2); i++)
						for (unsigned byte_offset = min_copy; byte_offset <= max_copy; byte_offset++)
							SetCompression(com, POSOV4, byte_offset, i, 0);
			}
		}
	}

	if (ENAINF) printf("\nNumber of all compressions: %u for size limits: %u %u  bits: %u\n", compressions, input_size, offset_limit, maxbits);
	if (!compressions) { printf("\nError: Not valid compression found.\n"); exit(1); }
	if (compressions > PACKTYPES) { printf("\nError: Number of compression types %u overflows %u for input size %u.\n", compressions, PACKTYPES, input_size); exit(1); }
	if (ENAINF) putchar('\n');
}

unsigned * TstAlloc(unsigned num)
{
	// Allocating memory with check of success //

	unsigned * memory = (unsigned*)malloc(num * sizeof(unsigned));
	if (memory) return memory;
	printf("Error: malloc(%u) returns null\n", num * (unsigned) sizeof(unsigned));
	exit(9); 
}

void DisplayPackModel(unsigned comsel, unsigned enable_check, const char * pass_name)
{
	// Recomputes total length and tax of packed data //
	// Optional: Displays pack model like LzxList     //
	// and checks integrity and optionality of model  //

	if (ENADSP || (ENASTA && compressions == 1))
	{
		printf("\nPack model for ... ");
		if (compressions > 1) DisplayCompression(comsel);
		puts(pass_name);
		putchar('\n');
	}
	unsigned kdesme = 0;	// Index to unpack data
	unsigned locove = 0;	// Local overhead
	unsigned loctax = 0;	// Local tax
	unsigned totpck = 0;	// Sum of all packed bytes
	unsigned totunp = 0;	// Sum of all unpack bytes
	unsigned totove = 0;	// Sum of all overhead
	unsigned tottax = 0;	// Total tax of pack model
	unsigned lstoff = 0;	// Last used offset for 2+ sekvences
	unsigned lstkde = 0;	// Last used offset position
	unsigned numsek = 0;	// Number of 2+ sekvences with own offset
	unsigned numreu = 0;	// Number of 2+ sekvences with reused offset sekvences
	unsigned numone = 0;	// Number of copied bytes

	if (ena_first_byte[typcom[comsel]])
	{
		kdesme = 1;
		totunp = 1;
		tottax = 8;
		if (ENADSP && !ENAZAC) printf("      +0..0       0000..0000    Tax:%6u                    First byte                        Data: %02X\n", tottax, *input_data);
	}

	while (kdesme < input_size)
	{
		unsigned length = buflen[kdesme];
		unsigned offset = bufofs[kdesme];

		if (!length)
		{
			printf("Error in pack model at  %+d  Incorrect zero length. Offset = %d\n", kdesme, offset);
			break;
		}

		if (kdesme + length > input_size)
		{
			printf("Error in pack model at  %u-%u  Input size %u overflow\n", kdesme, kdesme + length - 1, input_size);
			break;
		}

		if (length != buflen[kdesme + length - 1]
		||  offset != bufofs[kdesme + length - 1])
			printf("Error in pack model at  %u-%u  Inconsistent length: %u <> %u   offset: %u <> %u\n",
				kdesme, kdesme + length - 1,
				length, buflen[kdesme + length - 1],
				offset, bufofs[kdesme + length - 1]);

		if (offset)
		{
			unsigned tax_undo = 0;

			// Packed data //
			if (offset < 0xFFFF)
			{
				// Normal sekvence or copied byte //
				if (length > 1) numsek++; else numone++;
				locove = GetSekvenceTax(comsel, offset, length);

				// Check using of reused offset //
				if (enable_check && length > 1 && ena_rep_offset[typcom[comsel]])
				{
					unsigned preofs = bufofs[kdesme - 1];
					unsigned prelen = buflen[kdesme - 1];

					if (lstoff == offset
						&& (!preofs || prelen == 1))
					{
						unsigned pretax = (preofs
							? GetSekvenceTax(comsel, preofs, prelen)
							: unptax[typcom[comsel]][prelen]);
						unsigned oldtax = pretax + locove;
						unsigned reptax = GetUnpReusedOffsetSekTax(comsel, prelen, preofs, length);

						if (oldtax >= reptax)
						{
							printf("Error in pack model: Not used Reused offset %-5u from "
								"%u-%u len:%-5u to "
								"%u-%u len:%-5u "
								"PreOfs:%-5u PreLen:%-5u "
								"OldTax:%5u+%-5u = %-5u RepTax: %-5u Benefit: %c%d\n",
								lstoff,
								lstkde, lstkde + buflen[lstkde] - 1, buflen[lstkde],
								kdesme, kdesme + length - 1, length,
								preofs, prelen,
								pretax, locove, oldtax, reptax,
								oldtax > reptax ? '+':' ',
								oldtax - reptax);
						}
					}
					lstoff = offset;
					lstkde = kdesme;
				}
			}
			else
			{
				// Reused offset //
				numreu++;
				if (!numsek)
					printf("Error in pack model at  %u-%u  Reused offset: %u  len: %u  without previous normal sekvence !\n",
						kdesme, kdesme + length - 1, offset, length);

				unsigned preofs = bufofs[kdesme - 1];
				unsigned prelen = buflen[kdesme - 1];
				if (preofs && prelen > 1)
					printf("Error in pack model at  %u-%u  Reused offset: %u  Len: %u  after normal sekvence  ofs:%u  len:%u\n",
						kdesme, kdesme, offset, length, preofs, prelen);

				tax_undo = preofs
					? GetSekvenceTax(comsel, preofs, prelen)
					: unptax[typcom[comsel]][prelen];

				locove = GetUnpReusedOffsetSekTax(comsel, prelen, preofs, length);
			}

			if (ENADSP && (int)kdesme + (int)length >= ENAZAC && kdesme <= ENAEND)
				printf("%+8d..%-7u %04X..%04X    Tax:%6u + %-5u => %-6u  %s  Len:%-5u Ofs:%-5d   ",
					kdesme, kdesme + length - 1, kdesme, kdesme + length - 1,
					tottax - tax_undo, locove, tottax - tax_undo + locove,
					length > 1 ? (offset != 0xFFFF ? "Sekvence" : "Reused  ") : "CopyByte",
					length, offset);

			tottax += locove - tax_undo;
			totove += locove - tax_undo;
			totpck += length;
		}
		else
		{
			// Unpacked data //
			loctax = unptax[typcom[comsel]][length];
			locove = loctax - (length << 3);

			if (enable_check
				&& kdesme + length < input_size
				&& !bufofs[kdesme + length])
			{
				unsigned nxtlen = buflen[kdesme + length];
				unsigned nxttax = unptax[typcom[comsel]][nxtlen];
				unsigned sumtax = unptax[typcom[comsel]][nxtlen + length];
				unsigned kde3rd = kdesme + length + nxtlen;

				if (kde3rd < input_size)
				{
					if (bufofs[kde3rd] == 65535)
					{
						unsigned replen = buflen[kdesme + length + nxtlen];
						nxttax = GetUnpReusedOffsetSekTax(comsel, nxtlen, 0, replen);
						sumtax = GetUnpReusedOffsetSekTax(comsel, length + nxtlen, 0, replen);
					}
					else if (!bufofs[kde3rd])
						printf("Error in pack model: Wrong unpacked blocks at  %u-%u  %u-%u  %u-%u  Len: %u %u %u\n",
							kdesme, kdesme + length - 1,
							kdesme + length, kdesme + length + nxtlen - 1,
							kde3rd, kde3rd + buflen[kde3rd] - 1,
							length, nxtlen, buflen[kde3rd]);
				}

				if (loctax + nxttax >= sumtax)
					printf("Error in pack model: Wrong unpacked blocks at  %u-%u  and  %u-%u  Len: %u %u  Tax: %u + %u = %u >= %u\n",
						kdesme, kdesme + length - 1,
						kdesme + length, kdesme + length + nxtlen - 1,
						length, nxtlen, loctax, nxttax, loctax + nxttax, sumtax);
			}

			if (ENADSP && (int)kdesme + (int)length >= ENAZAC && kdesme <= ENAEND)
			{
				printf("%+8d..%-7u %04X..%04X    Tax:%6u + %-5u => %-6u  Unpacked  Len:%-5u .........   ",
					kdesme, kdesme + length - 1, kdesme, kdesme + length - 1,
					tottax, locove + (length << 3), tottax + locove + (length << 3),
					length);
			}

			totunp += length;
			totove += locove;
			tottax += loctax;
		}

		if (ENADSP && (int)kdesme + (int)length >= ENAZAC && kdesme <= ENAEND)
		{
			unsigned i;
			printf("  Data:");
			for (i = 0; i < 16 && i < length; i++) printf(" %02X", input_data[kdesme + i]);
			if (i < length)printf(" ...");
			putchar('\n');
		}
		kdesme += length;
	}

	unsigned endtax = endmark_length[typcom[comsel]];

	if (ENADSP)
		printf("%+8d..%-7u %04X..%04X    Tax:%6u + %-5u => %-6u  End mark  Total tax: %u bits => %u bytes\n\n",
			kdesme, kdesme, kdesme, kdesme, tottax, endtax, tottax + endtax, tottax + endtax, (tottax + endtax + 7) >> 3);

	totove += endtax;
	tottax += endtax;

	sumpck[comsel] = totpck;
	sumunp[comsel] = totunp;
	sumove[comsel] = totove;
	sumall[comsel] = tottax;
	sumsek[comsel] = numsek + numreu + numone;

	if (ENADSP || (ENASTA && compressions == 1))
	{
		if (compressions > 1) { printf("Compression: ");  DisplayCompression(comsel); putchar('\n');}

		printf("  Total packed length ....... %u (%u bits)\n", totunp + ((totove + 7) >> 3), (totunp << 3) + totove);
		printf("  Total unpack length ....... %u\n", input_size);
		printf("  Number of sekvences ....... %u\n", numsek);
		printf("  Number of reused offs ..... %u\n", numreu);
		printf("  Number of copied bytes .... %u\n", numone);
		printf("  Summary of packed bytes ... %u\n", totpck);
		printf("  Summary of unpack bytes ... %u\n", totunp);
		printf("  Compresion overhead ....... %u (%u bits)\n\n", (totove + 7) >> 3, totove);
	}
}

void OptimizeUnpacked(unsigned comsel, unsigned kdezac)
{
	// Search all 1 byte sekvences = one copied bytes in unpacked data block //
	// Result can be no optimal => next optimize unpacked blocks is needed   //

	unsigned totlen = buflen[kdezac];
	unsigned kdeend = kdezac + totlen;	// End of unpacked data block
	unsigned kdeunp = kdezac;	// Begin of new unpack block
	unsigned maxofs = 1u << bytcop[comsel];	// Maximal possible offset

	if (maxofs > offset_limit) maxofs = offset_limit;

	if (OPTREP > 3 && (int)kdezac + (int)totlen >= ENAZAC && kdezac <= ENAEND)
		printf("     Optimize unpacked    Processing block %+6d..%+-6d  Len: %-5u Max.offset: %u\n",
			kdezac, kdezac + totlen - 1, totlen, maxofs);

	for (unsigned kdesom = kdezac; kdesom < kdeend; kdesom++)
	{
		unsigned offset = bytofs[kdesom];

		if (offset <= maxofs)
		{
			unsigned lenunp = kdesom - kdeunp;
			if (lenunp)
			{
				if (OPTREP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
					printf("     Optimize unpacked %8u  Unpack block  Len: %u\n", kdeunp, lenunp);

				bufofs[kdeunp + 0] = 0;
				bufofs[kdesom - 1] = 0;
				buflen[kdeunp + 0] = lenunp;
				buflen[kdesom - 1] = lenunp;
			}
			if (OPTREP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
				printf("     Optimize unpacked %8u  [%02X]  Create one copied byte from %-5u Offset: %u\n",
					kdesom, input_data[kdesom], kdesom - offset, offset);

			kdeunp = kdesom + 1;
			buflen[kdesom] = 1;
			bufofs[kdesom] = offset;
			opti_one_cnt++;
		}
		else if (OPTREP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
			printf("     Optimize unpacked %8u  [%02X]  Offset: %u\n", kdesom, input_data[kdesom], offset);
	}
	if (kdezac < kdeunp && kdeunp < kdeend)
	{
		unsigned lenunp = kdeend - kdeunp;
		if (OPTREP > 3 && (int)kdeunp + (int)lenunp >= ENAZAC && kdeunp <= ENAEND)
			printf("     Optimize unpacked %8u UnpLast block  Len: %u\n", kdeunp, lenunp);

		bufofs[kdeunp + 0] = 0;
		bufofs[kdeend - 1] = 0;
		buflen[kdeunp + 0] = lenunp;
		buflen[kdeend - 1] = lenunp;
	}
}

void OptimizeReusedOffset(unsigned comsel, unsigned setmod, unsigned * unp_len_tax, unsigned * sek_len_tax,
						  unsigned kdezac, unsigned kdeend, unsigned * set_kde_end, unsigned * set_benefit)
{
	// Optimizes sekvences with reused offset - switch offset to reused and delete sekvences in conflict with examined offset //

	// Input:
	// setmod = 0 ... Only measure pack model part - how many bits we can save by turn offsets to reused in sekvences. kdezac = begin of part. //
	// setmod > 0 ... Remove all sekvences with different offset and turn to reused offset all sekvences with the same offset as first sekvence in this part //

	// Output:
	// benkde = *set_kde_end ... 0 = no benefit found, > 0 found at interval kdezac..benkde
	// benmax = *set_benefit ... Value of found benefit if bende > 0

	unsigned zaclen = buflen[kdezac];	// Length of first 2+ sekvence
	unsigned zacofs = bufofs[kdezac];	// Offset of first 2+ sekvence
	unsigned kdesme = kdezac + zaclen;	// Work index for evaluating
	unsigned lstofs = zacofs;			// Offset of last found sekvence
	unsigned tmpofs = zacofs;			// Offset of processed data item
	unsigned tmplen = zaclen;			// Length of processed data item
	unsigned preofs = 0;	// Offset of previous data item
	unsigned prelen = 0;	// Length of previous data item
	/**/ int benmax = 0;	// Maximal value of positive benefit of optimizing
	unsigned benkde = 0;	// End of part where is reached the maximal benefit
	unsigned oldtax = 0;	// Total old tax to evaluate benefit
	unsigned newtax = 0;	// Total new tax to evaluate benefit
	unsigned tmptax = 0;	// Temporary new tax in case of new created unpacked block
	unsigned unplen = 0;	// Length of created unpack data block
	unsigned taxone = max_1byte_sekv[typcom[comsel]] ? taxlen[comsel][0] + bytcop[comsel] : 0; // Tax for one byte sekvences

	if (OPTREP > 3 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
	{
		if (OPTREP > 4 || setmod)
			printf("\n%s %6u..%-6d  TaxOne: %u\n"
				"    Init tax at   %6u..%-6d  Len: %-5u Ofs: %-5u\n",
				setmod ? "Set reused offset" : "Evaluate benefit ",
				kdezac, kdeend - 1, taxone,
				kdezac, kdesme - 1, zaclen, zacofs);
	}

	while (kdesme < kdeend && newtax < oldtax + 64)
	{
		preofs = tmpofs;
		prelen = tmplen;
		tmpofs = bufofs[kdesme];
		tmplen = buflen[kdesme];
		kdesme += tmplen;

		if (!tmpofs)
		{
			// Case unpack block //
			oldtax += unp_len_tax[tmplen];
			newtax += unp_len_tax[tmplen];
			tmptax += unp_len_tax[tmplen + unplen] - unp_len_tax[unplen];
			unplen += tmplen;

			if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u  %3u Unp => Unp %-3u  "
					"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u TmpTax: %-5u  Benefit: %+d %+d\n",
					kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
					unp_len_tax[tmplen], unp_len_tax[unplen],
					lstofs, unplen, oldtax, newtax, tmptax, oldtax - newtax, oldtax - tmptax);
		}
		else if (tmplen == 1)
		{
			// Case one copied byte //
			tmptax = newtax;
			unplen = 0;

			if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u      One => One      "
					"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u TmpTax: %-5u  Benefit: %+-3d\n",
					kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
					lstofs, unplen, oldtax, newtax, tmptax, oldtax - newtax);
		}
		else if (tmpofs == 65535)
		{
			// Case: reused offset //
			if (lstofs == zacofs)
			{
				// Reused offset after the same offset //
				tmptax = newtax;
				unplen = 0;

				if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u      Rep => Rep      "
						"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u TmpTax: %-5u  Benefit: %+-3d\n",
						kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
						lstofs, unplen, oldtax, newtax, tmptax, oldtax - newtax);
			}
			else
			{
				// Reused offset after another offset = must be removed //
				oldtax -= preofs ? taxone : unp_len_tax[prelen];
				oldtax += GetUnpReusedOffsetSekTax(comsel, prelen, preofs, tmplen);
				tmptax += unp_len_tax[tmplen + unplen] - unp_len_tax[unplen];
				unplen += tmplen;

				if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u  %3u Rep => Unp %-3u  "
						"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u TmpTax: %-5u  Benefit: %+-3d\n",
						kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
						GetUnpReusedOffsetSekTax(comsel, prelen, preofs, tmplen), unp_len_tax[unplen],
						lstofs, unplen, oldtax, newtax, tmptax, oldtax - tmptax);

				if (setmod)
				{
					bufofs[kdesme - tmplen] = 0;
					bufofs[kdesme - 1] = 0;
					opti_del_cnt++;

					if (taxone) OptimizeUnpacked(comsel, kdesme - tmplen);
				}

				while (kdesme < kdeend && !bufofs[kdesme])
				{
					// Include following unpacked blocks into estimation //

					tmplen = buflen[kdesme];
					oldtax += unp_len_tax[tmplen];
					tmptax += unp_len_tax[tmplen + unplen] - unp_len_tax[unplen];
					unplen += tmplen;
					kdesme += tmplen;

					if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: 0      %3u Unp -> Unp %-3u  "
							"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u               TmpTax: %-5u  Benefit: %+-3d\n",
							kdesme - tmplen, kdesme - 1, tmplen,
							unp_len_tax[tmplen], unp_len_tax[unplen],
							lstofs, unplen, oldtax, tmptax, oldtax - tmptax);
				}
				newtax = tmptax;
				tmpofs = 0;
			}
		}
		else
		{
			// Case: Sekvence +2 with own offset //
			lstofs = tmpofs;
			/**/ int bentmp = oldtax - tmptax;
			oldtax += sek_len_tax[tmplen] + ofstax[tmpofs];

			if (tmpofs != zacofs)
			{
				// Sekvence has another offset = must be removed //
				tmptax += unp_len_tax[tmplen + unplen] - unp_len_tax[unplen];
				unplen += tmplen;

				if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u  %3u Sek => Unp %-3u  "
						"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u               TmpTax: %-5u  Benefit: %+-3d\n",
						kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
						sek_len_tax[tmplen] + ofstax[tmpofs], unp_len_tax[unplen],
						lstofs, unplen, oldtax, tmptax, oldtax - tmptax);

				if (setmod)
				{
					bufofs[kdesme - tmplen] = 0;
					bufofs[kdesme - 1] = 0;
					opti_del_cnt++;

					if (taxone) OptimizeUnpacked(comsel, kdesme - tmplen);
				}

				while (kdesme < kdeend && !bufofs[kdesme])
				{
					// Include following unpacked blocks into estimation //
					tmplen = buflen[kdesme];
					oldtax += unp_len_tax[tmplen];
					tmptax += unp_len_tax[tmplen + unplen] - unp_len_tax[unplen];
					unplen += tmplen;
					kdesme += tmplen;

					if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: 0      %3u Unp -> Unp %-3u  "
							"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u               TmpTax: %-5u  Benefit: %+-3d\n",
							kdesme - tmplen, kdesme - 1, tmplen,
							unp_len_tax[tmplen], unp_len_tax[unplen],
							lstofs, unplen, oldtax, tmptax, oldtax - tmptax);
				}
				newtax = tmptax;
				tmpofs = 0;
			}
			else
			{
				// Sekvence has the same offset = can be switched to reused offset //
				unsigned kdesek = kdesme - tmplen;
				unsigned sekcnt = 1;

				while (bufofs[kdesme] == zacofs && buflen[kdesme] >= 2)		// Collect more sekvences with the same offset //
				{
					if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					{
						newtax += sek_len_tax[tmplen] + ofstax[tmpofs];
						printf("   Check sek at   %6u..%-6d  Len: %-5u Ofs: %-5u  %3u Sek -> Rep %-3u  "
							"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u SekLen: %-5u  SekCnt: %u\n",
							kdesme - tmplen, kdesme - 1, tmplen, zacofs,
							sek_len_tax[tmplen] + ofstax[tmpofs], sek_len_tax[kdesme - kdesek] + ofstax[tmpofs],
							lstofs, unplen, oldtax, newtax, kdesme - kdesek, sekcnt);
					}

					tmplen = buflen[kdesme];
					oldtax += sek_len_tax[tmplen] + ofstax[tmpofs];
					kdesme += tmplen;
					sekcnt++;
				}

				tmplen = kdesme - kdesek; // Summary length of all sekvences with the same offset //

				if (unplen)
				{
					// Previous data is (or will be turned to) unpacked block //
					tmptax += GetUnpReusedOffsetSekTax(comsel, unplen, 0, tmplen) - unp_len_tax[unplen];
				}
				else if (preofs && prelen == 1)
				{
					// Previous data is one copied byte //
					tmptax += GetUnpReusedOffsetSekTax(comsel, 1, 1, tmplen) - taxone;
				}
				else
				{
					// Any other previos data (another sekvence or repeated offset //
					printf("\nError in benefit estimation: Wrong sekvence at  %u-%u  Len:%u Ofs:%u  after  %u-%u  Len:%u Ofs:%u  UnpLen: %u\n",
						kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
						kdesek - prelen, kdesek - 1, prelen, preofs,
						unplen);
					break;
				}

				if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				{
					newtax += sek_len_tax[tmplen] + ofstax[tmpofs];
					printf("   Check tax at   %6u..%-6d  Len: %-5u Ofs: %-5u  %3u Sek => Rep %-3u  "
						"LstOfs: %-5u UnpLen: %-5u OldTax: %-5u NewTax: %-5u TmpTax: %-5u  Benefit: %+-3d\n",
						kdesme - tmplen, kdesme - 1, tmplen, tmpofs,
						sek_len_tax[tmplen] + ofstax[tmpofs],
						unplen
							? GetUnpReusedOffsetSekTax(comsel, unplen, 0, tmplen)
							: GetUnpReusedOffsetSekTax(comsel, 1, 1, tmplen),
						lstofs, unplen, oldtax, newtax, tmptax, oldtax - tmptax);
				}

				unplen = 0;
				tmpofs = 65535;
				newtax = tmptax;

				if (setmod)
				{
					buflen[kdesme - tmplen] = tmplen;
					bufofs[kdesme - tmplen] = 65535;
					buflen[kdesme - 1] = tmplen;
					bufofs[kdesme - 1] = 65535;
					opti_use_cnt++;
				}

				int benact = oldtax - newtax;

				if (benact < bentmp)
				{
					if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("Break to lower benefit %+d => %+d\n", bentmp, benact);
					kdesme = kdesek;
					break;
				}

				if (benact >= benmax)
				{
					benmax = benact;
					benkde = kdesme;

					if (OPTREP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("  Set new max benefit:  BenMax: %+-5d BenKde: %-5u\n", benmax, benkde);
				}
			}
		}

		if (!tmplen)
		{
			printf("Error in optimize at  %+d  Incorrect zero length. Offset = %d\n", kdesme, tmpofs);
			break;
		}
	}

	*set_kde_end = benkde;
	*set_benefit = benkde ? benmax : 0;
}

void DumpDatabase(unsigned comsel, unsigned source, unsigned sekdel)
{
	// Dump pack database for debugging purpuses //

	// source = 0 ... select database: memofs memlen
	// source = 1 ... select database: bufofs buflen
	// sekdel = 1 ... show delimiters between blocks

	const char * separate = "  =================================================================================";

	printf("\nDump pack database for ");
	DisplayCompression(comsel);
	puts("\n");

	unsigned sekcnt = 0;

	for (unsigned kdesme = ENAZAC; kdesme <= ENAEND && kdesme < input_size; kdesme++)
	{
		if (sekdel)
		{

			if (!sekcnt && (source ? bufofs[kdesme] : memofs[kdesme][comsel]))
			{
				sekcnt = source ? buflen[kdesme] : memlen[kdesme][comsel];
				if (sekcnt && !(source ? bufofs[kdesme - 1] : memofs[kdesme - 1][comsel])) puts(separate);
			}
			else if (kdesme == 1 && ena_first_byte[typcom[comsel]]) puts(separate);
		}

		unsigned char byte = input_data[kdesme];
		printf("%+8d %04X   Data: 0x%02X  [%c]  ",
			kdesme, kdesme, byte,
			byte >= 32 && byte <= 126 ? byte : ' ');

		if (bytofs[kdesme] < 65536)
			printf("%5u", bytofs[kdesme]);
		else
			printf("     ");

			printf("    Length: %-5u  Offset: %-5u  Tax: %u\n",
			source ? buflen[kdesme] : memlen[kdesme][comsel],
			source ? bufofs[kdesme] : memofs[kdesme][comsel],
			memtax[kdesme][comsel]);

		if (sekcnt) { sekcnt--; if (!sekcnt && kdesme < input_size - 1) puts(separate); }
	}
	putchar('\n');
}

int main(int number_of_params, char **parameters)
{
	puts("\nBusy soft: Universal LZX Packer 02\n");

	if (number_of_params < 2)
	{
		puts(
			"Usage: LzxPack filename <options>\n\n"
			"Options:\n"
			" -s ............ Show statistics of all compressions used (default: do not show)\n"
			" -a ............ Save all compressed files (default: save only the best result)\n"
			" -i ............ Generate include file with compression type and parameters\n"
			" -o <file> ..... Set output file (default: input file + compression type + .lzx)\n"
			" -d <file> ..... File with depacker lengths to incorporate them into statistics\n"
			" -l <limit> .... Select maximum dictionary size (default: no size limit)\n"
			" -e <number> ... Select additonal effort to improve compression ratio (default: 0)\n"
			" -tnXYcNoAoB ... Select desired compression (default: 'n' and trying all options)\n"
			" -trXYcNoAoB ... The 'r' instead of 'n' forces reverse direction of depacking\n\n"
			"Selecting compression:\n"
			"  X ... compression type: 1=BLX 2=BLC 3=ZX9 4=ZX8 5=BS2 6=BX1 7=BX2 8=BX3 9=SX1\n"
			"  Y .... offset encoding: 1=OF1 2=OF2 3=OF3 4=OF4       6=OV1 7=OV2 8=OV3 9=OV4\n"
			"  N ... byte match offset bit-width used in BLX ZX9 BX1 BX2 BX3 SX1\n"
			"  A .... 1st match offset bit-width used in OF1 OF2 OF3 OF4 OV2 OV3 OV4\n"
			"  B .... 2nd match offset bit-width used in OF2 and OV3 only\n"
			"Omission of parameters X Y N A B (or zero) means trying all possible values.\n"
		);
		exit(1);
	}

	int time_start = (int)clock();
	int time_flag, time_set, time_now, time_prep_lz, time_lz_opti, time_opti_stat, time_stat_out, time_finish;

	//////////////////////////////////
	// Read command line parameters //
	//////////////////////////////////

	unsigned make_all_compressions = 0;
	unsigned make_include_file = 0;
	unsigned show_statistics = 0;
	unsigned set_output_name = 0;
	unsigned set_indent = 0;
	unsigned set_effort = 0;
	char *nam_inp = NULL;
	char *nam_dec = NULL;
	offset_limit = 65535;
	depackers = 0;

	compress_type = 0;
	compress_posi = 0;
	offset1bits = 0;
	offset2bits = 0;
	reverse_order = 0;

	nam_out[sizeof(nam_out) - 16] = 0;

	for (int i = 1; i < number_of_params; i++)
	{
		if (parameters[i][0] == '-')
		{
			char * xx = parameters[i] + 1;
			switch (*xx | 0x20)
			{
			case 'a': make_all_compressions = 1; break;
			case 'i': make_include_file = 1; break;
			case 's': show_statistics = 1; break;

			case 'o':
				xx++;
				if (!*xx)
				{
					if (i < number_of_params - 1) { i++; xx = parameters[i]; }
					else { puts("Error: No filename for -o option"); exit(1); }
				}
				strncpy(nam_out, xx, sizeof(nam_out) - 16);
				set_output_name = 1;
				break;

			case 'e':
				xx++;
				if (!*xx)
				{
					if (i < number_of_params - 1) { i++; xx = parameters[i]; }
					else { puts("Error: No number for -e option"); exit(1); }
				}
				set_effort = atoi(xx);
				if (set_effort > 65535) { puts("Error: Bad value in -e option"); exit(1); }
				break;


			case 'l':
				xx++;
				if (!*xx)
				{
					if (i < number_of_params - 1) { i++; xx = parameters[i]; }
					else { puts("Error: No number for -l option"); exit(1); }
				}
				offset_limit = atoi(xx);
				if (!offset_limit) { puts("Error: Dictionary size can not be zero"); exit(1); }
				if (offset_limit > 65535) offset_limit = 65535;
				break;

			case 't':
			{
				xx++;
				if (!*xx) { puts("Error: Compression type -tXX not selected"); exit(1); }
				if ((*xx | 0x20) == 'n') xx++;
				if ((*xx | 0x20) == 'r') { reverse_order = 1; xx++; }
				if (!*xx) break;
				if (*xx >= '0' && *xx <= '9') compress_type = *xx - '0';
				else { printf("Error: Unknown compression type '%c'\n", *xx); exit(1); }
				xx++;
				if (!*xx) break;
				if (*xx == '0' || GetNumberOffsets(*xx - '0') >= 0) compress_posi = *xx - '0';
				else { printf("Error: Unknown coding offset '%c'\n", *xx); exit(1); }
				xx++;
				if ((*xx | 0x20) == 'c')
				{
					xx++;
					if (*xx >= '0' && *xx <= '9') offset_copy = *xx - '0';
					else { printf("Error: Wrong offset bits number for one copied byte '%c'\n", *xx); exit(1); }
					xx++;
				}
				if ((*xx | 0x20) == 'o')
				{
					while (*xx)
					{
						xx++;
						if (*xx <'0' || *xx > '9') break;
						offset1bits = 10 * offset1bits + *xx - '0';
					}

					if ((*xx | 0x20) == 'o')
					{
						while (*xx)
						{
							xx++;
							if (*xx <'0' || *xx > '9') break;
							offset2bits = 10 * offset2bits + *xx - '0';
						}
					}
				}
				if (*xx) { printf("Error: Unexpected characters in type selection: %s\n", xx); exit(1); }
				if (offset1bits > 16 || offset2bits > 16) { puts("Error: Offset overflows maximal value 16"); exit(1); }
				if (offset2bits && (offset2bits <= offset1bits || offset2bits <= 2))
				{
					printf("Error: 2nd offset %u must be bigger than %s\n", offset2bits, offset2bits <= 2 ? "2" : "1st offset"); exit(1);
				}
				break;
			}

			case 'd':
			{
				xx++;
				if (!*xx)
				{
					if (i < number_of_params - 1) { i++; xx = parameters[i]; }
					else { puts("Error: No filename for -d option"); exit(1); }
				}
				nam_dec = xx;
				depackers = 0;
				FILE *file_inp = fopen(nam_dec, "rt"); if (!file_inp) { perror(nam_dec); exit(1); }
				memset(buffer, 0x00, sizeof(buffer));
				while (fgets(buffer, sizeof(buffer), file_inp))
				{
					buffer[1023] = 0;
					if (buffer[0] >= '1' && buffer[0] <= '9' && buffer[1] >= '1' && buffer[1] <= '9')
					{
						if (depackers < DEPACKERS)
						{
							depacker_com[depackers] = buffer[0] - '0';
							depacker_pos[depackers] = buffer[1] - '0';

							char * buf = buffer + 2;
							while(*buf == 0x20) buf++;
							depacker_len[depackers][0] = atoi(buf);
							while(*buf > 0x20) buf++;
							while(*buf == 0x20) buf++;
							depacker_len[depackers][1] = atoi(buf);
						}
						depackers++;
					}
					buffer[3] = 0;
				}
				fclose(file_inp);
				if (!depackers) { printf("\nError: Not valid depacked length definition found in file %s\n", nam_dec); exit(1); }
				if (depackers > DEPACKERS) { printf("\nError: Number of depackers %u overflows %u.\n", depackers, DEPACKERS); exit(1); }
				break;
			}
			default: { printf("Error: Unknown option %s\n", parameters[i]); exit(1); }
			}
		}
		else
		{
			if (nam_inp) { printf("Error: Too many input files:  %s  %s  ...\n", nam_inp, parameters[i]); exit(1); }
			nam_inp = parameters[i];
			if (!set_output_name) strncpy(nam_out, nam_inp, sizeof(nam_out) - 16);
		}
	}

	if (!nam_inp) { puts("Error: Missing input file !"); exit(1); }

	if (set_effort) printf("Selected additional effort: %d\n", set_effort);
	if (offset_limit < 65534) printf("Selected maximal dictionary size: %d\n", offset_limit);

	if (compress_type || compress_posi || offset_copy || offset1bits || offset2bits || reverse_order)
	{
		GetCompressType(buffer, -1, compress_type, compress_posi, offset_copy, offset1bits, offset2bits);
		printf("Selected compression: %s\n", buffer);
	}

	if (depackers)
	{
		printf("Selected depackers file: %s  (found %u depackers)\n", nam_dec, depackers);
		if (show_statistics)
		{
			putchar('\n');
			for (unsigned i = 0; i < depackers; i++)
				printf("  Compress: %c%c ... depacker length: %u\n",
					depacker_com[i] + '0', depacker_pos[i] + '0', depacker_len[i][reverse_order]);
		}
	}

	if (set_effort || compress_type || compress_posi || offset_copy || offset1bits || offset2bits || reverse_order || depackers) putchar('\n');

	////////////////////////
	// Process file names //
	////////////////////////

	unsigned namout_size = strlen(nam_out);
	char last_char = nam_out[namout_size - 1];
	if (last_char =='/' || last_char =='\\' || last_char ==':')
		{
		char * txt;
		for(txt = nam_inp + strlen(nam_inp); txt > nam_inp; txt--) if (*txt == '/' || *txt == '\\' || *txt == ':') break;
		if (txt > nam_inp) txt++;
		strcpy(nam_out + namout_size, txt);
		set_output_name = 0;
		}

	strcpy(ext_out, set_output_name ? "" : ".lzx");
	nam_out[sizeof(nam_out) - 1] = 0;
	for (char *txt = nam_out + strlen(nam_out);
		txt > nam_out && *txt != '/' && *txt != '\\';
		txt--)
		if (*txt == '.')
		{
			if (set_output_name) strcpy(ext_out, txt);
			*txt = 0;
			break;
		}

	int inp_len = strlen(nam_inp);
	int nam_len = strlen(nam_out);
	int out_len = strlen(nam_out) + strlen(ext_out) + (!set_output_name || make_all_compressions ? 14 : 0);
	set_indent = (inp_len < out_len ? out_len : inp_len);

	/////////////////////
	// Read input file //
	/////////////////////

	FILE *file_inp = fopen(nam_inp, "rb"); if (!file_inp) { perror(nam_inp); exit(1); }
	input_size = fread(input_data, 1, MAXLEN, file_inp);
	fclose(file_inp);

	if (reverse_order)
		for (int i = (input_size >> 1) - 1; i >= 0; i--)
		{
			unsigned char tmp = input_data[i];
			input_data[i] = input_data[input_size - i - 1];
			input_data[input_size - i - 1] = tmp;
		}

	printf("Load file: %-*s  Length: %u\n", set_indent, nam_inp, input_size);
	if (!input_size) { puts("Error: Zero length of input file !"); exit(1); }
	if (input_size < 2) { puts("Error: File is too short for compression !"); exit(1); }
	if (input_size == MAXLEN) { puts("Warning: File is too long. Only 65535 bytes will be compressed."); input_size = MAXLEN - 1; }

	////////////////////////////////////////
	// Prepare and check all compressions //
	////////////////////////////////////////

	PrepareAllCompressionTypes(input_size);

	if (ENAINF > 1)	// Display taxes and limits for offsets and lengths of sekvences //
	{
		#define VALS 39
		unsigned values[VALS];
		unsigned widths[VALS];
		for (int idx = 0, val = 1; idx < VALS - 1; idx++, val = val < 18 ? val + 1 : 3 * val >> 1)
		{
			values[idx] = val;
			widths[idx] = val < 1000 ? 4 : val < 10000 ? 5 : 6;
		}
		values[VALS - 1] = 65534;
		widths[VALS - 1] = 6;

		for (int s = 0; s < 3; s++)	// 0 = offsets, 1 = pack lengths, 2 = unpack lengths //
		{
			printf("Test for limits: %s             ", !s ? "Offsets      " : s == 1 ? "Pack length  " : "Unpack length");
			for (int idx = 0; idx < VALS; idx++) printf("%*u", widths[idx], values[idx]);
			printf("\n==========================================    ");
			for (int i = 1; i < 172; i++) putchar('='); putchar('\n');
			for (unsigned com = 0; com < compressions; com++)
			{
				DisplayCompression(com);
				char skip = s == 1 && !max_1byte_sekv[typcom[com]];
				if (skip) printf("   -");
				unsigned *limits = s ? limlen[com] : limoff[com];
				for (int idx = skip ? 1 : 0; idx < VALS; idx++)
				{
					unsigned i, tax;
					switch (s)
					{
					case 0:
						for (i = 0; limits[i] && values[idx] > limits[i]; i++);
						tax = taxofs[com][i];
						break;
					case 1:	tax = lentax[typcom[com]][values[idx]]; break;
					case 2:	tax = unptax[typcom[com]][values[idx]] - (values[idx] << 3); break;
					default: tax = 999;
					}
					if (tax && tax < 65536)
						printf("%*u", widths[idx], tax);
					else
						printf("     -" + 6 - widths[idx]);
				}
				putchar('\n');
			}
			putchar('\n');
		}

		if (ENAINF > 2) for (unsigned com = 0; com < compressions; com++) // Test GetSekvenceTax() //
		{
			char skip = !max_1byte_sekv[typcom[com]];
			printf("Test GetSekvenceTax for: "); DisplayCompression(com); printf("\n\nLengths =>");
			for (int idx = 0; idx < VALS; idx++) printf("%*u", widths[idx], values[idx]);
			printf("\nOffsets  ,"); for (int i = 1; i < 175; i++) putchar('-'); putchar('\n');
			for (int off = 0; off < VALS; off++)
			{
				printf("%7u  |", values[off]);
				if (skip) printf("   -");
				for (int len = skip ? 1 : 0; len < VALS; len++)
					if (maxoff[com] < values[off] || maxlen[com] < values[len])
						printf("     -" + 6 - widths[len]);
					else
						printf("%*u", widths[len], GetSekvenceTax(com, values[off], values[len])); putchar('\n');
			}
			putchar('\n');
		}
	}

	////////////////////////////////
	// Main LZ compression engine //
	////////////////////////////////

	time_flag = 0;
	time_prep_lz = clock();
	time_set = time_prep_lz + DELAY1;

	unsigned cmps = 0;
	if (ENACOM) printf("Searching sekvences ...\n\n");

	unsigned char byte = input_data[0];
	if (ENACOM && ENAZAC == 0)
		printf("    +0 0000  Byte: 0x%02X '%c'  First byte\n",
			byte, byte >= 32 && byte <= 126 ? byte : ' ');

	// Set tax for first standalone byte //
	for (unsigned comsel = 0; comsel < compressions; comsel++)
	{
		if (ena_first_byte[typcom[comsel]])
		{
			memlen[0][comsel] = 0;
			memtax[0][comsel] = 8;
		}
		else
		{
			memlen[0][comsel] = 1;
			memtax[0][comsel] = unptax[typcom[comsel]][1];
		}
		memofs[0][comsel] = 0;

		if (ENAUNP && !ENAZAC)
		{
			printf("\t  ");
			DisplayCompression(comsel);
			printf("Set init unpack tax at +%-5u  Len: %-5u  Tax: %u\n",
				0, memlen[0][comsel], memtax[0][comsel]);
		}
	}

	// Initializations //
	memset(hash_bytes, 0x11, sizeof(hash_bytes));
	hash_bytes[input_data[0]] = 0;

	repper[0] = 0;
	replen[0] = 0;
	lstidx[0] = 0;
	lstlen[0] = 0;
	bufnxt[0] = 0;
	bytofs[0] = 65536;

	// Main loop for all next bytes from input file //
	for (unsigned kdesme = 1; kdesme < input_size; kdesme++)
	{
		// Create word hash for quick searching //
		unsigned hash_index = input_data[kdesme] | (input_data[kdesme - 1] << 8);
		unsigned locate = hash_words[hash_index];
		bufnxt[kdesme] = locate;
		hash_words[hash_index] = kdesme;

		if (ENACOM && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
		{
			unsigned max = 8;
			unsigned char byte = input_data[kdesme];
			printf("%+6d %04X  Byte: 0x%02X '%c'  Create hash: 0x%04X  ",
				kdesme, kdesme, byte,
				byte >= 32 && byte <= 126 ? byte : ' ',
				hash_index);
			unsigned nxt = bufnxt[kdesme];
			if (nxt) printf("  Next hashes at offsets:");
			while (nxt && max)
			{
				printf(" %u", nxt);
				nxt = bufnxt[nxt];
				max--;
			}
			if (nxt) printf(" ...");
			putchar('\n');
		}

		// Set initial unpack tax for new processed unpack byte //
		// memlen[kdesme][comsel] = counter of not packed bytes //
		for (unsigned comsel = 0; comsel < compressions; comsel++)
		{
			unsigned oldtax = 0;
			unsigned oldlen = 0;
			unsigned oldidx = kdesme - 1;

			if (!memofs[oldidx][comsel])
			{
				oldlen = memlen[oldidx][comsel];
				oldidx = kdesme - oldlen - 1;
			}
			if (kdesme > oldlen) oldtax = memtax[oldidx][comsel];

			memofs[kdesme][comsel] = 0;
			memlen[kdesme][comsel] = oldlen + 1;
			memtax[kdesme][comsel] = oldtax + unptax[typcom[comsel]][memlen[kdesme][comsel]];

			if (ENAUNP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
			{
				printf("\t  ");
				DisplayCompression(comsel);
				printf("Set init unpack tax at +%-5u  Len: %-5u  Tax: %-6u   %u + 8*%u + %u   OldIdx:%d  OldLen:%d\n",
					kdesme,
					memlen[kdesme][comsel],
					memtax[kdesme][comsel],
					oldtax, memlen[kdesme][comsel], unptax[typcom[comsel]][memlen[kdesme][comsel]] - (memlen[kdesme][comsel] << 3), oldidx, oldlen);
			}
		}

		// Search one byte sekvences //
		unsigned char bajt = input_data[kdesme];
		unsigned offset = kdesme - hash_bytes[bajt];
		hash_bytes[bajt] = kdesme;
		bytofs[kdesme] = offset;

		if (offset <= 256 && offset <= offset_limit)
		{
			if (ENAONE && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("\tOne byte: Match 0x%02X at %+d  Offset:%u\n", bajt, kdesme - offset, offset);
			for (unsigned comsel = 0; comsel < compressions; comsel++) if (max_1byte_sekv[typcom[comsel]] && offset <= maxone[comsel])
			{
				unsigned newtax = memtax[kdesme - 1][comsel] + taxlen[comsel][0] + bytcop[comsel];

				if (ENAONE && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				{
					printf("\t  ");
					DisplayCompression(comsel);
					printf("Compare  UnpTax: %u  PckTax: %u   (%u+%u)",
						memtax[kdesme][comsel], newtax, memtax[kdesme - 1][comsel], taxlen[comsel][0] + bytcop[comsel]);
				}

				if (PCKALL || newtax < memtax[kdesme][comsel])
				{
					if (ENAONE && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("  accepted\n");
					memofs[kdesme][comsel] = offset;
					memlen[kdesme][comsel] = 1;
					memtax[kdesme][comsel] = newtax;
				}
				else if (ENAONE && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("  rejected\n");
			}
		}

		memset(locskp, 0x00, compressions * sizeof(unsigned));
		memset(idxoff, 0x00, compressions * sizeof(unsigned));
		memset(idxlen, 0x00, compressions * sizeof(unsigned));

		unsigned needed = 1;

		unsigned best_offset = 0;
		unsigned best_length = 0;
		unsigned repe_length = 0;
		unsigned repe_period = 0;
		unsigned maxloc = MAXLEN;

		// Loop for search all sekvences related to processed byte //
		while (locate && needed)
		{
			unsigned offset = kdesme - locate;

			if (locate < best_length)
			{
				if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("%+6d %04X                      "
						"Skip hash%+6d %04X   Offset:%-5d  "
						" and all next hashes due to bestlen = %u\n",
						kdesme, kdesme, locate, locate, offset, best_length);
				break;
			}

			if (locate > maxloc)
			{
				if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("%+6d %04X                      "
						"Skip hash%+6d %04X   Offset:%-5d  "
						"due to maxloc = %u\n",
						kdesme, kdesme, locate, locate, offset, maxloc);

				locate = bufnxt[locate];
				continue;
			}

			if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("%+6d %04X                      "
					"Test hash%+6d %04X   Offset:%-5d  "
					".....    RepPer:%-3u RepLen:%-5u    "
					"Previous is LstIdx:%-5u LstLen:%-5u \n",
					kdesme, kdesme, locate, locate, offset,
					repper[locate], replen[locate],
					lstidx[locate - 1], lstlen[locate - 1]);

			unsigned length = 0;
			///////////////////////////////////////////////////////
			if (lstidx[locate - 1] == kdesme - 1)
			{
				length = lstlen[locate - 1] + 1;

				if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("%+6d %04X                        "
						"Previous sekvence at %+d  Len:%u  =>  Set length:%u\n",
						kdesme, kdesme, kdesme - 1, lstlen[locate - 1], length);
			}
			///////////////////////////////////////////////////////
			else
			{
				unsigned maxlen = locate + 1;
				unsigned period = repper[locate];
				unsigned subsub = kdesme - period;
				if (period && period == repper[subsub])
				{
					if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("%+6d %04X                        "
							"Periodic sekvences detected:  "
							"%+d Period:%u Length:%u   and  "
							"%+d Period:%u Length:%u\n",
							kdesme, kdesme, locate, period, replen[locate],
							subsub, repper[subsub], replen[subsub]);

					unsigned limit = period << 1; if (limit > maxlen) limit = maxlen;
					for (length = 2; length < limit; length++)
					{
						if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
							printf("%+6d %04X                          "
								"Compare before %+6d 0x%02X   %+6d 0x%02X   Len:%-5u \n",
								kdesme, kdesme,
								locate - length, input_data[locate - length],
								kdesme - length, input_data[kdesme - length],
								length);

						if (ENACOM) cmps++;
						if (input_data[locate - length] != input_data[kdesme - length]) break;
					}
					if (length == limit)
					{
						length = period + replen[subsub];
						if (length > replen[locate]) length = replen[locate];
						if (length > locate + 1) length = locate + 1;
						maxloc = kdesme - length + 1;
						if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
							printf("%+6d %04X                        "
								"Increase length %u to %u   Maxloc: %u\n",
								kdesme, kdesme, limit, length, maxloc);

						while (length < maxlen)
						{
							if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("%+6d %04X                          "
									"Compare after  %+6d 0x%02X   %+6d 0x%02X   Len:%-5u \n",
									kdesme, kdesme,
									locate - length, input_data[locate - length],
									kdesme - length, input_data[kdesme - length],
									length);

							if (ENACOM) cmps++;
							if (input_data[locate - length] != input_data[kdesme - length]) break;
							length++;
						}
					}
				}

				else for (length = 2; length < maxlen; length++)
				{
					if (ENACOM > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("%+6d %04X                          "
							"Compare normal %+6d 0x%02X   %+6d 0x%02X   Len:%-5u\n",
							kdesme, kdesme,
							locate - length, input_data[locate - length],
							kdesme - length, input_data[kdesme - length],
							length);

					if (ENACOM) cmps++;
					if (input_data[locate - length] != input_data[kdesme - length]) break;
				}
			}
			/////////////////////////
			lstidx[locate] = kdesme;
			lstlen[locate] = length;
			/////////////////////////
			if (ENACOM > 1 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("\tFound sekvence ... "
					"Loc:%-+5d  Off:%-5d  Len:%-5u    RepPer:%-3u RepLen:%-5u\n",
					locate, offset, length, repper[locate], replen[locate]);

			if (replen[locate])
			{
				unsigned tmpper = repper[locate];
				unsigned tmplen = replen[locate]; if (tmplen > length) tmplen = length;
				if (tmplen && (tmplen - tmpper > repe_length - repe_period) && (tmplen > tmpper))
				{
					repe_period = tmpper;
					repe_length = tmplen;
				}
			}

			if (best_length < length && offset <= offset_limit)
			{
				if (ENACOM > 1 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("\tProcess sekvence ... OldLen:%u  NewLen:%u  NewOfs:%u  Locate:+%u\n",
						best_length, length, offset, locate);

				best_length = length;
				best_offset = offset;

				// Selecting and accepting sekvences for each compression //
				needed = 0;
				for (unsigned comsel = 0; comsel < compressions; comsel++)
				{
					if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					{
						printf("\t");
						DisplayCompression(comsel);
						printf("Processing sekvence  Ofs:%u  Len:%u ... ", offset, length);
					}
					if (locskp[comsel])
					{
						if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Skipped\n");
						continue;
					}
					if (offset > maxoff[comsel])
					{
						locskp[comsel] = 1;
						if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Skipped: Offset %u > MaxOff %u\n", offset, maxoff[comsel]);
						continue;
					}

					needed = 1;

					if (memofs[kdesme][comsel] && length + 2 < memlen[kdesme][comsel])
					{
						if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Skipped: Length %u + 2 < BufLen %u\n", length, memlen[kdesme][comsel]);
						continue;
					}

					if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("(maxoff: %u)\n", maxoff[comsel]);

					// Find offset tax //
					if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("\t  Find tax for offset %u ...", offset);
					int seloff = idxoff[comsel];
					if (offset > limoff[comsel][seloff])
					{
						while (limoff[comsel][seloff] && offset > limoff[comsel][seloff])
						{
							if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" %u", limoff[comsel][seloff]);
							seloff++;
						}

						if (!limoff[comsel][seloff])
						{
							locskp[comsel] = 1;
							if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" Skipped: Offset %u > LimOff %u\n", offset, limoff[comsel][seloff - 1]);
							continue;
						}

						idxoff[comsel] = seloff;
					}
					unsigned offtax = taxofs[comsel][seloff];
					if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" found:%u tax:%u\n", limoff[comsel][seloff], taxofs[comsel][seloff]);

					// Find and check length //
					unsigned bstoff = 99999;
					unsigned bstlen = 0;
					unsigned bstidx = 0;
					unsigned bsttax = memtax[kdesme][comsel];

					if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("\t  Find best sek for  MaxLen:%u  OldTax:%u\n", length, bsttax);

					int sellen = idxlen[comsel];
					if (limlen[comsel][sellen] == 1) sellen++;	// Only 2+ byte sekvences here //

					while (limlen[comsel][sellen])
					{
						unsigned lentax = taxlen[comsel][sellen];
						unsigned cmptax = bsttax - offtax - lentax;
						unsigned tmplen = limlen[comsel][sellen];
						unsigned tstend = tmplen; if (tstend > length) tstend = length;
						unsigned tstzac = limlen[comsel][sellen - 1] + 1;

						if (tstzac < 2) tstzac = 2;
						if (set_effort < tstend - tstzac) tstzac = tstend - set_effort; // <= speed up according to effort //

						if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("\t  sellen:%u  cmptax:%u  Len:%u  Eff:%u  Int: %u-%u => %u-%u\n",
							sellen, cmptax, length, set_effort, limlen[comsel][sellen - 1] + 1, limlen[comsel][sellen], tstzac, tstend);

						for (unsigned tstlen = tstend; tstlen >= tstzac; tstlen--)
						{
							// Check normal sekvences //
							unsigned oldtax = memtax[kdesme - tstlen][comsel];

							if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("\t    Check len:%-5u SekTax:%2u+%-3u  PreKde:%-5u  PreTax:%-6u =>  NewTax: %u",
									tstlen, lentax, offtax, kdesme - tstlen, oldtax, offtax + lentax + oldtax);

							if (PCKALL
								|| oldtax < cmptax
								|| (oldtax == cmptax && tstlen > bstlen))
							{
								cmptax = oldtax;
								bstoff = offset;
								bstlen = tstlen;
								bstidx = sellen;
								bsttax = cmptax + offtax + lentax;
								if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("    BstTax: %u ... accepted\n", bsttax);
							}
							else if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("    BstTax: %u\n", bsttax);
						}

						if (length <= tmplen)
						{
							if (length == tmplen) sellen++;
							break;
						}

						sellen++;
					}
					idxlen[comsel] = sellen;

					// Check for repeated offset //
					// for full sequences only ! //
					if (ena_rep_offset[typcom[comsel]])
					{
						unsigned kdebck = kdesme - length;
						unsigned bckoff = memofs[kdebck][comsel];
						unsigned bcklen = memlen[kdebck][comsel];

						if (!bckoff ||
							(bckoff && bcklen == 1))
						{
							unsigned repbck = kdebck - bcklen;

							if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("\t  Reused offset ... Start search  Repkde:%u  Repoff:%u  Kdebck:%u  Repbck:%u  for  sekvence Ofs:%u Len:%u\n",
									repkde[comsel], repoff[comsel], kdebck, repbck, offset, length);

							unsigned preofs = 0;
							unsigned kdeofs = kdebck;

							// Find bookmark or previous sekvence with normal offset //
							while (kdeofs > 0)
							{
								if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
									printf("\t  Reused offset ...   Find rep offset ... Kdeofs:%-5u  Offs:%-5u  Len:%-5u\n",
										kdeofs, memofs[kdeofs][comsel], memlen[kdeofs][comsel]);

								if (memlen[kdeofs][comsel] > 1 && memofs[kdeofs][comsel] > 0 && memofs[kdeofs][comsel] < 0xFFFF)
								{
									// Found sekvence with regular offset //
									preofs = memofs[kdeofs][comsel];
									if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
										printf("\t  Reused offset ...   Found sekvence  Offset: %u  Length: %u\n", preofs, memlen[kdeofs][comsel]);
									break;
								}
								if (kdeofs == repkde[comsel])
								{
									// Found bookmark with known offset of some previous sekvence //
									if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
										printf("\t  Reused offset ...   Found bookmark  Offset: %u\n", repoff[comsel]);
									preofs = repoff[comsel];
									break;
								}
								kdeofs -= memlen[kdeofs][comsel];
							}

							if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("\t  Reused offset ... Kdebck: %u  Kdeofs: %u   PrevLen: %u  PrevOfs: %u  ThisOfs: %d %s\n",
									kdebck, kdeofs, memlen[kdeofs][comsel], preofs, offset,
									preofs == offset ? "... match" : "(no)");

							if (repbck > 0)
							{
								// Set new bookmark for future //
								repkde[comsel] = repbck; // -memlen[repbck][comsel];
								repoff[comsel] = preofs;

								if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
									printf("\t  Reused offset ... Set bookmark  Repkde=%u  Repoff=%u\n", repkde[comsel], repoff[comsel]);
							}

							if (preofs == offset)
							{
								// Size of unpacked data before sekvence //
								unsigned reptax = GetUnpReusedOffsetSekTax(comsel, bcklen, bckoff, length);
								unsigned unstax = memtax[kdebck - bcklen][comsel] + reptax;

								if (ENAREP && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
									printf("\t  Reused offset ... Unp/Byt %u-%u:%u  Sek: %u-%u:%u  Tax: %u    TotalTax: %u + %u = %u   (OldTax:%u)%s\n",
										kdebck - bcklen + 1, kdebck, bcklen,
										kdesme - length + 1, kdesme, length,
										reptax, memtax[kdebck][comsel], reptax, unstax, bsttax,
										unstax < bsttax ? " Accepted" : "");

								if (PCKALL || unstax < bsttax)
								{
									bstoff = 0xFFFF;
									bsttax = unstax;
									bstlen = length;
								}
							}
						}
					}

					if (bstlen)
					{
						if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
							printf("\t  Accept sekvence ... Length: %u  Offset: %u  OldTax: %u  NewTax: %u\n",
								bstlen, bstoff, memtax[kdesme][comsel], bsttax);

						memofs[kdesme][comsel] = bstoff;
						memlen[kdesme][comsel] = bstlen;
						memtax[kdesme][comsel] = bsttax;

						if (bstlen == maxlen[comsel])
						{
							locskp[comsel] = 1;
							if (ENACPT && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("\t  Skip all next matches due to maximal possible length %u\n", maxlen[comsel]);
						}
					}
				}
				/////////////////////////////////////////////////////////////
			}

			if (ENACOM > 3 && !needed && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("Skip all next hashes (not needed to find next sekvences)\n");

			locate = bufnxt[locate];
		}
		if (best_length)
		{
			if (best_length > best_offset)
			{
				repper[kdesme] = best_offset;
				replen[kdesme] = best_length + best_offset;
			}
			else
			{
				if (repe_length > best_length) repe_length = best_length;
				if (repe_length <= repe_period) { repe_length = 0; repe_period = 0; }
				repper[kdesme] = repe_period;
				replen[kdesme] = repe_length;
			}

			if (ENACOM > 3 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
				printf("\t  Best sekvence ... Locate:%-+5d  Offset:%-5d  Length:%-5u    RepPer:%-3u RepLen:%-5u   Cmps:%u\n",
					kdesme - best_offset, best_offset, best_length, repper[kdesme], replen[kdesme], cmps);
		}

		if (ENASUM && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
		{
			if (compressions > 1) printf("\tSummmary of best sekvences\n");
			for (unsigned comsel = 0; comsel < compressions; comsel++)
			{
				if (compressions == 1)
					printf("\tBest sekvence ... ");
				else
				{
					printf("\t  ");
					DisplayCompression(comsel);
				}
				printf("Kdesme: %-5u  TotalTax: %-5u  ", kdesme, memtax[kdesme][comsel]);
				if (memofs[kdesme][comsel])
					printf("Offset: %-5u  Length: %-5u\n", memofs[kdesme][comsel], memlen[kdesme][comsel]);
				else
					puts("none");
			}
		}
		if (SEKNUM >= 0)
			printf(SEKNAM "  kdesme:%-5u  Offset:%-5u  Length:%-5u  Tax: %-5u  Dif: %d\n",
				kdesme, memofs[kdesme][SEKNUM], memlen[kdesme][SEKNUM],
				memtax[kdesme][SEKNUM], memtax[kdesme][SEKNUM] - memtax[kdesme - 1][SEKNUM]);

		// Show time progress //
		time_now = clock();
		if (time_now - time_set > 0)
		{
			time_flag = 1;
			time_set = time_now + DELAY2;
			int percent = 100 * kdesme / input_size;
			fprintf(stderr, "   Compress %2u %%\r", percent);
		}
	}
	if (time_flag) fprintf(stderr, "\r                    \r");

	if (ENACOM && input_size <= ENAEND) printf("%+6d %04X  End of file\n\n", input_size, input_size);

	// End of main LZ compression engine //

	if (OPTREP == 1 || OPTUNP == 1) putchar('\n');

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	opti_com_cnt = 0;	// Number of compressions changed by optimization process //

	time_flag = 0;
	time_lz_opti = clock();
	time_set = time_lz_opti + DELAY1;

	// Begin of main loop for all used compressions //
	for (unsigned comsel = 0; comsel < compressions; comsel++)
	{
		if (ENADMP) DumpDatabase(comsel, 0, 0);

		unsigned taxone = taxlen[comsel][0] + bytcop[comsel];	// Tax of one copied byte

		unsigned enable_reused_offset
					   = ena_rep_offset[typcom[comsel]];	// Flag if reused offset can be used in processed compression //
		unsigned begin = ena_first_byte[typcom[comsel]];	// Begin of pack model in destination data = can be 0 or 1    //

		unsigned * unp_len_tax = unptax[typcom[comsel]];	// None packed blocks length taxes table for actual used COMxxx //
		unsigned * sek_len_tax = lentax[typcom[comsel]];	// Sekvences 2+ bytes length taxes table for actual used COMxxx //

		// Create offset taxes table for actual used POSxxx //
		{
			ofstax[0] = 990000000;
			unsigned idx = 0;
			unsigned fil = 1;
			while (1)
			{
				unsigned lim = limoff[comsel][idx];
				if (!lim) break;
				if (lim > input_size) lim = input_size;
				unsigned tax = taxofs[comsel][idx];
				if (ENAINF > 3) printf("Set offset tax  Idx:%-2u Range: %5u-%-5u Cnt:%-5u Tax:%-2u\n", idx, fil, lim, lim - fil + 1, tax);
				for (; fil <= lim; fil++) ofstax[fil] = tax;
				if (fil > input_size) break;
				idx++;
			}
			if (fil < input_size) memset(ofstax + fil, 0x77, (input_size - fil) * sizeof(unsigned));
		}

		if (ENAINF > 1)
		{
			printf("\nOffset tax overhead  (bits:min-max) for ");
			DisplayCompression(comsel);
			putchar('\n');

			if (ENAINF > 3) for (int i = 0; i < 256; i++) printf(" %u", ofstax[i]); putchar('\n');

			unsigned max = 28;
			unsigned idx = 0;

			while (max && idx <= input_size)
			{
				unsigned nxt = idx;
				unsigned bits = ofstax[nxt];
				while (nxt <= input_size && bits == ofstax[nxt]) nxt++;
				printf("  %d:%u-%u", bits, idx, nxt - 1);
				idx = nxt;
				max--;
			}
			if (!max && idx <= input_size) printf("  ...");
			putchar('\n');
		}

		/////////////////////////////////////////////////
		// Read and unreverse pack model from database //
		/////////////////////////////////////////////////

		if (ENAREV)
		{
			printf("\nUnreverse sekvences ");
			if (compressions > 1) DisplayCompression(comsel);
			printf(" in range %u - %u\n\n", begin, input_size - 1);
		}

		buflen[0] = 0;
		bufofs[0] = 0;

		unsigned kdesme = input_size - 1;
		while (1)
		{
			if (kdesme > 65535) { printf("Error: Inconsistent pack model in unreverse: Com: %u  Ptr: %d\n", comsel, kdesme); exit(16); }

			unsigned length = memlen[kdesme][comsel];
			unsigned offset = memofs[kdesme][comsel];

			if (!length || length > kdesme - begin + 1) { printf("Error: Incorrect length at %+d   Com: %u  Offset: %d  Length: %d\n", kdesme, typcom[comsel], offset, length); exit(15); }

			buflen[kdesme] = length;
			bufofs[kdesme] = offset;

			kdesme -= length - 1;

			buflen[kdesme] = length;
			bufofs[kdesme] = offset;

			if (ENADMP) for (unsigned i = 1; i < length - 1; i++)
			{
				buflen[kdesme + i] = length;
				bufofs[kdesme + i] = offset;
			}

			if (offset)
			{
				if (ENAREV && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("Unreverse sekvence  %5u..%-5u  Len:%-5u  Off:%+d \n",
						kdesme, kdesme + length - 1, length, offset);

				if (kdesme && offset == 0xFFFF && memofs[kdesme - 1][comsel])
				{
					unsigned tusme = kdesme - 1;
					unsigned preofs = memofs[tusme][comsel];
					unsigned prelen = memlen[tusme][comsel];
					if (prelen > 1)
					{
						printf("Error: ");
						if (compressions > 1) DisplayCompression(comsel);
						printf("Reused offset %d at %u-%u:%u after normal sekvence offset %d at %u-%u:%u  (read)\n",
							offset, kdesme, kdesme + length - 1, length,
							preofs, kdesme - prelen, tusme, prelen);
					}
				}
			}
			else
			{
				if (ENAREV && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("Unreverse unpacked  %5u..%-5u  Len:%-5u \n",
						kdesme, kdesme + length - 1, length);
			}

			if (kdesme == begin) break;
			kdesme--;
		}

		if (ENADMP) DumpDatabase(comsel, 1, 1);

		if (ENASTA)
		{
			unsigned one_byte = 0;
			unsigned stat_max_own = 0;
			unsigned stat_max_rep = 0;
			unsigned stat_low_own = 0;
			unsigned stat_low_rep = 0;
			unsigned stat_unp_rep = 0;
			unsigned stat_pck_rep = 0;

			kdesme = begin;
			while (kdesme < input_size)
			{
				unsigned len = buflen[kdesme];
				unsigned ofs = bufofs[kdesme];

				if (ofs && len == 1) one_byte++;

				if (ofs == 0xFFFF)
				{
					if (bufofs[kdesme - 1])
						stat_pck_rep++;
					else
						stat_unp_rep++;
				}

				if (ofs && len > 1)
				{
					if (kdesme > ofs + 1 && (input_data[kdesme - 1] == input_data[kdesme - ofs - 1]))
					{
						if (ofs < 0xFFFF) stat_low_own++; else stat_low_rep++;
					}
					else
					{
						if (ofs < 0xFFFF) stat_max_own++; else stat_max_rep++;
					}
				}
				kdesme += len;
			}

			if (compressions > 1) DisplayCompression(comsel);
			printf("OneByte:%-5u MaxSek:%-5u LowSek:%-4u MaxRep:%-3u LowRep:%-3u UnpRep: %-3u PckRep:%u\n",
				one_byte, stat_max_own, stat_low_own, stat_max_rep, stat_low_rep, stat_unp_rep, stat_pck_rep);
		}

		// Compute length with end marks //
		sumcom[comsel] = memtax[input_size - 1][comsel] + endmark_length[typcom[comsel]];
		sumall[comsel] = 0;

		DisplayPackModel(comsel, 0, "LZ compression");

		if (sumall[comsel] != sumcom[comsel])
		{
			printf("Error: Bad length (bits)   Estimated: %-6u  Measured: %-6u  Dif: %+-5d  in  ", sumcom[comsel], sumall[comsel], sumall[comsel] - sumcom[comsel]);
			DisplayCompression(comsel);
			putchar('\n');
		}

		/////////////////////////
		// Optimize pack model //
		/////////////////////////

		if (OPTREP > 2 || OPTUNP > 2)
		{
			printf("\nOptimize pack model for ");
			DisplayCompression(comsel);
			printf("LZ compress length: %u %u\n\n", (sumall[comsel] + 7) >> 3, sumall[comsel]);
		}

		opti_pas_cnt = 0;

		unsigned optimize_length = 0;	// Used for check if optimize process makes packed data bigger //
		unsigned optimize_pass = 1;

		if (PCKALL < 2) do	// Start of main optimize loop //
		{
			opti_rem_tax = 0;
			opti_rem_cnt = 0;
			opti_use_cnt = 0;
			opti_del_cnt = 0;
			opti_one_cnt = 0;
			opti_use_tax = 0;

			if (use_blk_unpack[typcom[comsel]])
			{
				if (OPTUNP > 2) printf("\nOptimize unpacked blocks - remove sekvences\n\n");

				optimize_length = sumall[comsel];
				kdesme = begin;			// Global index to pack model //

				while (kdesme < input_size)
				{
					//  Skip all sekvences 4+ bytes longer  //
					while (kdesme < input_size && bufofs[kdesme] && buflen[kdesme] >= 4)
					{
						if (OPTUNP > 4 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
							printf("  Skip    %6u..%-6u Sekvence   Len: %-5u Ofs: %u\n",
								kdesme, kdesme + buflen[kdesme] - 1, buflen[kdesme], bufofs[kdesme]);

						kdesme += buflen[kdesme];
					}
					if (kdesme >= input_size) break;

					unsigned reuidx = 0;		// Where is the next sekvence - used for reused offset checking
					unsigned reustx = 0;		// Status for necesity of keeping 2+ sekvence due to next sekvence with reused offset
					unsigned sumite = 0;		// Counter for number of item   in actual examined area
					unsigned sumtax = 0;		// Increment bit length counter
					unsigned tmptax = 0;		// Temporary bit length counter
					unsigned optidx = 0;		// Index to optzac/optend/optben
					unsigned kdesom = kdesme;	// Index to examined area in optimization process
					unsigned tstzac = kdesme;	// Begin of examined area in optimization process

					buftax[kdesme] = sumtax;	// Incremental tax of actual pack model

					while (kdesom < input_size)
					{
						const char * item = "unknown";
						unsigned tmplen = buflen[kdesom];
						unsigned tmpofs = bufofs[kdesom];

						if (tmpofs)
						{
							if (tmplen == 1)
							{
								if (OPTUNP) item = "Copied byte  ";
								tmptax = taxone;
							}
							else if (tmpofs == 65535)
							{
								if (OPTUNP) item = "Reused offset";
								unsigned prelen = buflen[kdesom - 1];
								unsigned preofs = bufofs[kdesom - 1];
								tmptax = GetUnpReusedOffsetSekTax(comsel, prelen, preofs, tmplen) - (preofs ? taxone : unp_len_tax[prelen]);
							}
							else
							{
								if (OPTUNP) item = "Sekvence ... ";
								tmptax = sek_len_tax[tmplen] + ofstax[tmpofs];
							}
						}
						else
						{
							if (OPTUNP) item = "Unpacked ... ";
							tmptax = unp_len_tax[tmplen];
						}

						if (OPTUNP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
							printf("SetTax    %6u..%-6u Len: %-5u Ofs: %-5u %s Tax: %6u + %-3u => %-6u  Items: %u\n",
								kdesom, kdesom + tmplen - 1, tmplen, tmpofs, item, sumtax, tmptax, sumtax + tmptax, sumite + 1);

						kdesom += tmplen;
						sumtax += tmptax;
						sumite++;
						buftax[kdesom] = sumtax;

						if (tmpofs && tmplen >= 4) break;	// Sekvence length 4+ is end of search local area

						if (enable_reused_offset && reuidx < kdesom)
						{
							// Find following 2+ sekvence to determine if it has reused offset //
							reustx = 0; // No reused offset found in next item //
							reuidx = kdesom;
							while (reuidx < input_size && (!bufofs[reuidx] || buflen[reuidx] < 2))
							{
								if (OPTUNP > 4 && (int)reuidx >= ENAZAC && reuidx <= ENAEND)
									printf("  SeaSek  %6u..%-6u Len: %-5u Ofs: %u\n",
										reuidx, reuidx + buflen[reuidx] - 1, buflen[reuidx], bufofs[reuidx]);

								reuidx += buflen[reuidx];
							}
							if (reuidx < input_size && bufofs[reuidx] == 65535) reustx = 1;

							if (OPTUNP > 3 && (int)reuidx >= ENAZAC && reuidx <= ENAEND)
								printf("  SeaSek  %6u..%-6u Len: %-5u Ofs: %-5u    Reused status set to %u\n",
									reuidx, reuidx + buflen[reuidx] - 1, buflen[reuidx], bufofs[reuidx], reustx);
						}

						unsigned count = 16;			// Number of examined items (0 = all)
						unsigned reulen = 0;		// Length of following reused offset (0 = none)
						unsigned bstrat = 0;		// Ratio of unpack optimization
						unsigned bstzac = 0;		// Begin of unpack optimization with best ratio
						unsigned bstcnt = 0;		// Number of items with best ratio (0 = benefit not found)
						unsigned kdetmp = kdesom;	// Pointer to pack model
						unsigned endtax = sumtax;	// Tax at end of tested block

						if (sumite < 64 || !(sumite & 0xFF)) count = 64;

						if (enable_reused_offset && (bufofs[kdesom] == 65535))
						{
							// If reused offset follows, tax evaluate must be different //
							reulen = buflen[kdesom];
							endtax = buftax[kdesom - tmplen] + GetUnpReusedOffsetSekTax(comsel, tmplen, tmpofs, buflen[kdesom]);
							// endtax includes following sekvence with reused offset //

							if (OPTUNP > 3 && (int)kdesom >= ENAZAC && kdetmp <= ENAEND)
								printf("   Reused %6d..%-6u Len: %-5u Ofs: %-5u    Change EndTax %u => %u = %u + %u\n",
									kdesom, kdesom + tmplen - 1, reulen, bufofs[kdesom],
									sumtax, endtax, buftax[kdesom - tmplen], endtax - buftax[kdesom - tmplen]);
						}

						while (kdetmp > tstzac)
						{
							kdetmp = kdetmp - buflen[kdetmp - 1];

							if (reustx && buflen[kdetmp] >= 2 && bufofs[kdetmp] && bufofs[kdetmp] != 65535)
							{
								if (OPTUNP > 3 && (int)kdetmp >= ENAZAC && kdetmp <= ENAEND)
									printf("    Break %6d..%-6u Len: %-5u Ofs: %-5u    Reused offset follows at %u\n",
										kdetmp, kdesom - 1, buflen[kdetmp], bufofs[kdetmp], reuidx);
								break;
							}

							unsigned length = kdesom - kdetmp;
							unsigned oldtax = endtax - buftax[kdetmp];
							unsigned newtax = reulen
								? GetUnpReusedOffsetSekTax(comsel, length, 0, reulen)
								: unp_len_tax[length];

							int diftax = oldtax - newtax;
							if (diftax >= 0)
							{
								unsigned ratio = (diftax << 15) / length;
								if (ratio > 65535) ratio = 65535;

								if (bstrat <= ratio)
								{
									bstcnt = count;
									bstrat = ratio;
									bstzac = kdetmp;
								}
							}

							if (OPTUNP > 4 && (int)kdetmp >= ENAZAC && kdetmp <= ENAEND)
							{
								unsigned ratio = diftax <= 0 ? 0 : (diftax << 15) / length;
								printf("    Ratio %6d..%-6u Len: %-5u Ofs: %-5u Cnt: %-4u ReuLen: %-5u UnpLen: %-5u "
									" OldTax: %-5u NewTax: %-5u Benefit: %+-5d Ratio: %-5u"
									"   Bst: %6u..%-6d BstRat: %d\n",
									kdetmp, kdesom - 1, buflen[kdetmp], bufofs[kdetmp],
									count, reulen, length, oldtax, newtax, diftax, ratio,
									bstzac, kdesom - 1, bstrat);
							}

							count--; if (!count) break;
						}

						if (bstcnt
							&& (bufofs[bstzac] || buflen[bstzac] + bstzac < kdesom)) // Find ony one unpacked block is not accepted //
						{
							if (OPTUNP > 4 && (int)bstzac >= ENAZAC && bstzac <= ENAEND)
								printf("  Find    %6u..%-6u Len: %-5u ReuLen: %-5u Benefit: %-5d BstRat: %-5u\n",
									bstzac, kdesom - 1, kdesom - bstzac, reulen,
									reulen
									? buftax[kdesom - tmplen] + GetUnpReusedOffsetSekTax(comsel, tmplen, tmpofs, buflen[kdesom])
									- buftax[bstzac] - GetUnpReusedOffsetSekTax(comsel, kdesom - bstzac, 0, reulen)
									: sumtax - buftax[bstzac] - unp_len_tax[kdesom - bstzac],
									bstrat);

							optidx++;
							optzac[optidx] = bstzac;
							optend[optidx] = kdesom;
							optben[optidx] = bstrat;

							if (OPTUNP > 3 && (int)bstzac >= ENAZAC && bstzac <= ENAEND)
							{
								printf("SetNew %-2u %6u..%-6u Len: %-5u Ratio: %-5u          KdeTmp: %u TstZac: %u\n",
									optidx, optzac[optidx], optend[optidx] - 1, optend[optidx] - optzac[optidx], optben[optidx],
									kdetmp, tstzac);

								if (OPTUNP > 5)
								{
									putchar('\n');

									for (unsigned i = 1; i <= optidx; i++)
										printf("  List %-2u %6u..%-6u Len: %-5u Ratio: %-5u\n",
											i, optzac[i], optend[i] - 1, optend[i] - optzac[i], optben[i]);
								}
							}
						}

						if (OPTUNP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND) putchar('\n');
					}

					if (optidx)
					{
						if (OPTUNP > 2 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
							printf("\nSummary   %6u..%-6u Length: %-5u SumTax: %-5u %-6u Items: %-5u Clock: %u\n\n",
								kdesme, kdesom - 1, kdesom - kdesme, (sumtax + 7) >> 3, sumtax, sumite, (unsigned)(clock() - time_lz_opti));

						if (OPTUNP > 3 && (int)kdesom >= ENAZAC && kdesom <= ENAEND)
						{
							for (unsigned i = 1; i <= optidx; i++)
								printf("  List %-2u %6u..%-6u Len: %-5u Ratio: %-5u\n",
									i, optzac[i], optend[i] - 1, optend[i] - optzac[i], optben[i]);

							putchar('\n');
						}

						while (optidx)
						{
							unsigned remidx = optidx;
							unsigned ratio = optben[optidx];
							unsigned kdezac = optzac[optidx];
							unsigned kdeend = optend[optidx];
							unsigned length = kdeend - kdezac;

							if (OPTUNP > 3 && (int)kdeend >= ENAZAC && kdezac <= ENAEND)
							{
								printf(" Check %-3u %6u..%-6u Len: %-5u Ratio: %-5d\n",
									remidx, kdezac, kdeend - 1, length, ratio);
							}

							while (1)
							{
								// Check benefit of all previous findings intercurrent to removed items //
								optidx--;
								if (!optidx) break;
								if (optend[optidx] <= kdezac) break;

								if (OPTUNP > 3 && (int)kdeend >= ENAZAC && kdezac <= ENAEND)
									printf("  Test %-3u %6u..%-6u Len: %-5u Ratio: %-5u    check with %-2u %6u..%-6u Ratio: %-5u  %s\n",
										optidx, optzac[optidx], optend[optidx] - 1, optend[optidx] - optzac[optidx], optben[optidx],
										remidx, kdezac, kdeend - 1, ratio,
										optben[optidx] > ratio ? "better ratio found" : "");

								if (optben[optidx] > ratio) { length = 0; break; }

							}

							if (!length) continue;

							unsigned oldtax = buftax[kdeend] - buftax[kdezac];
							unsigned newtax = unp_len_tax[length];

							if (kdeend < input_size && bufofs[kdeend] == 65535)
							{
								unsigned replen = buflen[kdeend];
								newtax = GetUnpReusedOffsetSekTax(comsel, length, 0, replen);
								oldtax = GetUnpReusedOffsetSekTax(comsel, buflen[kdeend - 1], bufofs[kdeend - 1], replen)
									+ buftax[kdeend - buflen[kdeend - 1]]
									- buftax[kdezac];
							}

							bufofs[kdezac - 0] = 0;
							bufofs[kdeend - 1] = 0;
							buflen[kdezac - 0] = length;
							buflen[kdeend - 1] = length;

							opti_rem_cnt++;
							opti_rem_tax += oldtax - newtax;

							if (OPTUNP > 2 && (int)kdeend >= ENAZAC && kdezac <= ENAEND)
							{
								printf("Remove %-3u %6u..%-6u Len: %-5u Ratio: %-5u OldTax: %-5u NewTax: %-5u Benefit: %-5d BenSum: %u\n",
									remidx, kdezac, kdeend - 1, length, ratio, oldtax, newtax, oldtax - newtax, opti_rem_tax);

								if (OPTUNP > 3) putchar('\n');
							}
						}
					}
					else
					{
						if (OPTUNP > 4 && (int)kdesom >= ENAZAC && kdesme <= ENAEND)
							printf("\n  Skip    %6u..%-6u\n", kdesme, kdesom - 1);
						kdesme = kdesom;
					}
				}

				if (OPTUNP > 4) printf("   End    %6u..%-6u\n", kdesme, kdesme);

				if (OPTUNP > 2)
				{
					if (OPTUNP > 2) putchar('\n');
					printf("Optimize result: Create %u unpacked blocks. Benefit: %u bits = about %u bytes\n",
						opti_rem_cnt, opti_rem_tax, (opti_rem_tax + 4) >> 3);
				}

				if (opti_rem_cnt)
				{
					opti_pas_cnt++;
					DisplayPackModel(comsel, 0, "Optimized unpacked blocks");

					if (OPTUNP > 2)
					{
						unsigned new_bits = sumall[comsel];
						unsigned old_bytes = (optimize_length + 7) >> 3;
						unsigned new_bytes = (new_bits + 7) >> 3;
						printf("Optimize length change: %u - %u = %+d = %+d bits ... about %u - %u = %+d bytes\n",
							optimize_length, new_bits, optimize_length - new_bits, opti_rem_tax,
							old_bytes, new_bytes, old_bytes - new_bytes);
					}

					if (optimize_length < sumall[comsel])
						printf("Error in optimize process. Packed data length: %u %u => %u %u\n",
							(optimize_length + 7) >> 3, optimize_length, (sumall[comsel] + 7) >> 3, sumall[comsel]);

					if (optimize_length - opti_rem_tax != sumall[comsel])
						printf("Error in optimize process. Incorrect length old %u - benefit %u = new %u verzus %u\n",
							optimize_length, opti_rem_tax, optimize_length - opti_rem_tax, sumall[comsel]);

					optimize_length = sumall[comsel];
				}
			}						// End of optimize unpacked blocks //

			if (enable_reused_offset)
			{
				if (OPTREP > 2) printf("\nOptimize reused offsets - reconfigure sekvences\n\n");

				unsigned idx = 0;	// Index to optzac/optend/optben //
				kdesme = begin;
				while (kdesme < input_size)
				{
					unsigned length = buflen[kdesme];
					unsigned offset = bufofs[kdesme];

					if (offset && offset != 65535 && length >= 2)
					{
						unsigned benefit = 0;
						unsigned kdeend = 0;

						while(1)
						{
							unsigned nxtkde = kdesme + buflen[kdesme];

							if (bufofs[nxtkde] != offset) break;
							if (buflen[nxtkde] < 2) break;

							if (OPTREP > 3 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("Skip sekvence %6d..%-6d Len: %-5u Ofs: %-5u  because %6d..%-6d Len: %-5u Ofs: %-5u\n",
									kdesme, nxtkde - 1, buflen[kdesme], bufofs[kdesme],
									nxtkde, nxtkde + buflen[nxtkde] - 1, buflen[nxtkde], bufofs[nxtkde]);

							kdesme = nxtkde;
						}

						OptimizeReusedOffset(comsel, 0, unp_len_tax, sek_len_tax, kdesme, input_size, &kdeend, &benefit);

						if (kdeend)
						{
							if (OPTREP > 3 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("Benefit %4u  found in %6d..%-6d =>  Store %u\n", benefit, kdesme, kdeend, idx);

							optzac[idx] = kdesme;
							optend[idx] = kdeend;
							optben[idx] = benefit;
							idx++;
						}
					}
					kdesme += buflen[kdesme];
				}

				unsigned items = idx;

				if (OPTREP > 2)
				{
					printf("\nFound %u optimizable areas.\n\n", items);
					for (unsigned i = 0; i < items; i++)
						printf("Area %-4u for optimize: %8d..%-8d Len: %-5u Ofs: %-5u Benefit: %+d\n",
							i, optzac[i], optend[i], optend[i] - optzac[i] + 1, bufofs[optzac[i]], optben[i]);
					if (items) putchar('\n');
					if (OPTREP > 3 && items) printf("Execute reused offset optimizations\n");
				}

				while (items)
				{
					// Search maximal benefit //
					// In case of more the same benefits - the shortest part has selected //

					unsigned benidx = 0;
					unsigned benmax = optben[benidx];

					for (unsigned i = 1; i < items; i++)
						if (benmax < optben[i] || (benmax == optben[i] && optend[benidx] - optzac[benidx] > optend[i] - optzac[i]))
						{
							benidx = i;
							benmax = optben[i];
						}

					// Test for regular 2+ sekvence //
					unsigned kdezac = optzac[benidx];
					unsigned kdeend = optend[benidx];
					unsigned sekofs = bufofs[kdezac];

					if (OPTREP > 4 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
						printf("\nMaximum benefit found: Area %-4u %6u..%-6u Len: %-5u Ofs: %-5u Ben: %+-5d   Pack model %5u..%-5u Len: %-5u Ofs: %u\n",
							benidx, kdezac, kdeend, kdeend - kdezac, bufofs[optzac[benidx]], optben[benidx],
							kdezac, kdezac + buflen[kdezac] - 1, buflen[kdezac], sekofs);

					if (!sekofs || sekofs == 65535 || buflen[kdezac] < 2) // Area must begin by regular 2+ sekvence with own offset //
					{
						// Measured part begins with reused offset => it was modified => Remove this part from list //
						if (OPTREP > 2 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
							printf("Area %-4u  processing  %6u..%-6u  Ben: %+-5d =>  Already processed  Ofs: %u\n",
								benidx,
								kdezac, kdeend, benmax, sekofs);

						optend[benidx] = 0;
					}
					else
					{
						// Re-evaluate benefint in found maximum //
						unsigned tmpend = 0;
						unsigned tmpben = 0;
						OptimizeReusedOffset(comsel, 0, unp_len_tax, sek_len_tax, kdezac, kdeend, &tmpend, &tmpben);

						if (OPTREP > 4 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
							printf("Benefit %+3d found in %6u..%-6u\n", tmpben, kdezac, tmpend);

						if (benmax != tmpben || kdeend != tmpend)
						{
							// Measured part was modified by previous optimalization //
							if (OPTREP > 2 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
								printf("Area %-4u  processing  %6d..%-6d  Ben: %+-5d =>  Modified %6u..%-6u  Ben: %+d\n",
									benidx,
									kdezac, kdeend, benmax,
									kdezac, tmpend, tmpben);

							// if (tmpend)
							{
								optend[benidx] = tmpend;
								optben[benidx] = tmpben;
							}
						}
						else
						{
							// Previous and this measure is the same => we can execute the optimization //
							unsigned tmpend = 0;
							unsigned tmpben = 0;
							OptimizeReusedOffset(comsel, 1, unp_len_tax, sek_len_tax, kdezac, kdeend, &tmpend, &tmpben);

							opti_use_tax += tmpben;

							if (OPTREP > 2 && (int)kdezac >= ENAZAC && kdezac <= ENAEND)
								printf("Area %-4u  processing  %6u..%-6u  Ben: %+-5d =>  Optimized  Count: %u\n",
									benidx, kdezac, kdeend, benmax, opti_use_cnt);

							if (tmpend != kdeend || tmpben != benmax)
								printf("Error: Inconsistent benefit evaluation in area %u:  %u-%u  Old benefit: %+-5d => %u-%u  New benefit: %+d\n",
									benidx,
									kdezac, optend[benidx], benmax,
									kdezac, kdeend, tmpben);

							optend[benidx] = 0;
						}
					}
					if (!optend[benidx])
					{
						// Erase processed item from list //
						items--;
						if (benidx != items)
						{
							optzac[benidx] = optzac[items];
							optend[benidx] = optend[items];
							optben[benidx] = optben[items];
						}
					}
				}

				if (OPTREP > 2)
				{
					if (OPTREP > 2) putchar('\n');
					printf("Optimize result: Delete %u sek + Set %u reused offsets + Create %u copied bytes.\n",
						opti_del_cnt, opti_use_cnt, opti_one_cnt);
				}

				if (opti_del_cnt || opti_use_cnt || opti_one_cnt)
				{
					opti_pas_cnt++;
					DisplayPackModel(comsel, 0, "Optimized reused offset");
					optimize_length = sumall[comsel];
				}
			}								// End of optimize reused offset //

			if (OPTREP > 1 || OPTUNP > 1)
			{
				if (OPTREP > 2 || OPTUNP > 2) putchar('\n');
				if (compressions > 1) DisplayCompression(comsel);
				printf("Optimize pass:%-3u Rem:%-5u Del:%-5u Use:%-5u One:%-5u Clock: %u\n",
					optimize_pass, opti_rem_cnt, opti_del_cnt, opti_use_cnt, opti_one_cnt, (unsigned)(clock() - time_lz_opti));
			}
			optimize_pass++;

		}
		while (opti_del_cnt || opti_use_cnt || opti_one_cnt); // End of main optimize loop //

		// Some addtional final optimizations //

		opti_use_cnt = 0;
		opti_del_cnt = 0;

		if (enable_reused_offset)
		{
			if (OPTREP > 2) printf("\nOptimize reused offsets - final adding reused offsets\n\n");

			unsigned lstofs = 0;	// Last found offset
			kdesme = begin;			// Global index to pack model

			while (kdesme < input_size)
			{
				unsigned tmplen = buflen[kdesme];
				unsigned tmpofs = bufofs[kdesme];

				if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("%8u..%-6u Len: %-5u Ofs: %-5u ", kdesme, kdesme + tmplen - 1, tmplen, tmpofs);

				if (tmpofs && tmplen >= 2 && tmpofs != 65535)
				{
					if (tmpofs == lstofs)
					{
						unsigned preofs = bufofs[kdesme - 1];
						unsigned prelen = buflen[kdesme - 1];

						if (!preofs || prelen == 1)
						{
							unsigned pretax = (preofs
								? taxone
								: unptax[typcom[comsel]][prelen]);
							unsigned oldtax = pretax + sek_len_tax[tmplen] + ofstax[tmpofs];
							unsigned newtax = GetUnpReusedOffsetSekTax(comsel, prelen, preofs, tmplen);

							if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
								printf("OldTax: %-5u NewTax: %-5u Benefit: %c%-5d",
									oldtax, newtax,
									oldtax > newtax ? '+' : ' ',
									oldtax - newtax);

							if (oldtax >= newtax)
							{
								if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf(" Set reused offset");

								bufofs[kdesme] = 65535;
								bufofs[kdesme + tmplen - 1] = 65535;

								opti_use_cnt++;
							}
						}
						else
						{
							if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Sekvence after sekvence");
						}
					}
					else
					{
						if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Set lstofs");
						lstofs = tmpofs;
					}
				}

				kdesme += tmplen;
				if (OPTREP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) putchar('\n');
			}
		}

		if (use_bit_unpack[typcom[comsel]] && use_blk_unpack[typcom[comsel]])
		{
			if (OPTUNP > 2) printf("\nOptimize unpacked blocks - split to one byte and rest of block\n\n");

			kdesme = begin;

			while (kdesme < input_size)
			{
				unsigned tmplen = buflen[kdesme];
				unsigned tmpofs = bufofs[kdesme];

				if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("%8u..%-6u Len: %-5u Ofs: %-5u  ", kdesme, kdesme + tmplen - 1, tmplen, tmpofs);

				if (!tmpofs && tmplen > 2)
				{
					unsigned condition = 0;
					unsigned taxlen = tmplen;

					if (enable_reused_offset && bufofs[kdesme + tmplen] == 65535)
					{
						taxlen--;
						unsigned oldtax = GetUnpReusedOffsetSekTax(comsel, tmplen, 0, buflen[kdesme + tmplen]);
						unsigned newtax = GetUnpReusedOffsetSekTax(comsel, taxlen, 0, buflen[kdesme + tmplen]) + unp_len_tax[1];
						condition = newtax < oldtax;

						if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Reused");
					}
					else
					{
						condition = unp_len_tax[taxlen] > unp_len_tax[taxlen - 1] + unp_len_tax[1];
						if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) printf("Normal");
					}

					if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
						printf("  Tax: %5u => %u = %u + %-5u",
							unp_len_tax[taxlen], unp_len_tax[1] + unp_len_tax[taxlen - 1], unp_len_tax[1], unp_len_tax[taxlen - 1]);

					if (condition)
					{
						unsigned len = tmplen - 1;
						bufofs[kdesme + 1] = 0;
						buflen[kdesme + 0] = 1;
						buflen[kdesme + 1] = len;
						buflen[kdesme + len] = len;

						if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
							printf(" Benefit: %-5d  Split to 1 + %-5u",
								unp_len_tax[taxlen] - unp_len_tax[1] - unp_len_tax[taxlen - 1], len);

						opti_del_cnt++;
					}
				}
				if (OPTUNP > 2 && (int)kdesme >= ENAZAC && kdesme <= ENAEND) putchar('\n');

				kdesme += tmplen;
			}

			if (OPTUNP > 2) putchar('\n');
		}

		if (opti_pas_cnt || opti_use_cnt || opti_del_cnt)
		{
			if (ENADMP) DumpDatabase(comsel, 1, 1);
			DisplayPackModel(comsel, 1, "Final statistics");
			if (sumcom[comsel] != sumall[comsel]) opti_com_cnt++;
		}
		sumdec[comsel] = depackers && declen[comsel] ? ((sumall[comsel] + 7) >> 3) + declen[comsel] : 0;

		/////////////////////////////////////////
		// Write pack model back into database //
		/////////////////////////////////////////

		if (ENAREV)
		{
			printf("\nStore pack model ");
			if (compressions > 1) DisplayCompression(comsel);
			printf(" in range %u - %u\n\n", begin, input_size - 1);
		}

		for (unsigned kdesme = begin; kdesme < input_size;)
		{
			unsigned length = buflen[kdesme];
			unsigned offset = bufofs[kdesme];

			if (!length || kdesme + length > input_size) { printf("Error: Incorrect length at %+d   Com: %u  Offset: %d  Length: %d\n", kdesme, typcom[comsel], offset, length); exit(15); }

			memlen[kdesme][comsel] = length;
			memofs[kdesme][comsel] = offset;

			if (ENADMP) for (unsigned i = 1; i < length; i++)
			{
				memlen[kdesme + i][comsel] = length;
				memofs[kdesme + i][comsel] = offset;
			}

			if (offset)
			{
				if (ENAREV && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("Store sekvence  %5u..%-5u  Len:%-5u  Off:%+d \n",
						kdesme, kdesme + length - 1, length, offset);

				if (kdesme && offset == 0xFFFF && bufofs[kdesme - 1])
				{
					unsigned tusme = kdesme - 1;
					if (buflen[tusme] > 1)
					{
						printf("Error: ");
						if (compressions > 1) DisplayCompression(comsel);
						printf("Reused offset %d at %u-%u:%u after normal sekvence offset %d at %u-%u:%u  (write)\n",
							offset, kdesme, kdesme + length - 1, length,
							bufofs[tusme], kdesme - buflen[tusme], tusme, buflen[tusme]);
					}
				}
			}
			else
			{
				if (ENAREV && (int)kdesme >= ENAZAC && kdesme <= ENAEND)
					printf("Store unpacked  %5u..%-5u  Len:%-5u \n",
						kdesme, kdesme + length - 1, length);
			}

			kdesme += length;
		}

		if (ENADMP) DumpDatabase(comsel, 0, 1);

		// Show time progress //
		time_now = clock();
		if (time_now - time_set > 0)
		{
			time_flag = 1;
			time_set = time_now + DELAY2;
			int percent = 100 * comsel / compressions;
			fprintf(stderr, "   Optimize %2u %%\r", percent);
		}
	}

	if (time_flag) fprintf(stderr, "\r                    \r");

	////////////////////////////////
	// Find best and worst result //
	////////////////////////////////

	time_opti_stat = clock();

	int best_len_result = 99999999;
	int best_dec_result = 99999999;
	int wrst_len_result = -9999999;
	int wrst_dec_result = -9999999;
	int best_len_compress = -1;
	int best_dec_compress = -1;
	int wrst_len_compress = -1;
	int wrst_dec_compress = -1;
	for (unsigned i = 0; i < compressions; i++)
	{
		if (best_len_result >(int)sumall[i])
		{
			best_len_compress = i;
			best_len_result = (int)sumall[i];
		}
		if ((best_dec_result > (int)sumdec[i]) && declen[i])
		{
			best_dec_compress = i;
			best_dec_result = (int)sumdec[i];
		}
		if (wrst_len_result < (int)sumall[i])
		{
			wrst_len_compress = i;
			wrst_len_result = (int)sumall[i];
		}
		if ((wrst_dec_result < (int)sumdec[i]) && declen[i])
		{
			wrst_dec_compress = i;
			wrst_dec_result = (int)sumdec[i];
		}
	}

	if (ENAINF || ENASTA)
	{
		printf("\n  Best compress Idx: %-3d  Len: %d\n", best_len_compress, best_len_result);
		if (depackers) printf(  "  Best depacker Idx: %-3d  Len: %d\n", best_dec_compress, best_dec_result);
		printf(  " Worst compress Idx: %-3d  Len: %d\n", wrst_len_compress, wrst_len_result);
		if (depackers) printf(  " Worst depacker Idx: %-3d  Len: %d\n", wrst_dec_compress, wrst_dec_result);
		if (!OPTUNP || OPTREP) putchar('\n');
	}

	if ((OPTUNP || OPTREP) && opti_com_cnt)
	{
		printf("\nOptimized %u of %u compressions:\n\n", opti_com_cnt, compressions);
		for (unsigned i = 0; i < compressions; i++) if (OPTUNP > 1 || OPTREP > 1 || (sumcom[i] != sumall[i]))
		{
			unsigned bitlzc = sumcom[i];
			unsigned bitend = sumall[i];
			DisplayCompression(i);
			printf("Comp: %5u %6u", (bitlzc + 7) >> 3, bitlzc);
			if (bitend != bitlzc) printf("  Opti: %5u %6u  Dif: %5d %5d", (bitend + 7) >> 3, bitend, (bitlzc - bitend + 7) >> 3, bitlzc - bitend);
			if (bitend == (unsigned)best_len_result) printf("   ... best result");
			putchar('\n');
		}
		putchar('\n');
	}

	////////////////
	// Statistics //
	////////////////

	if (show_statistics)
	{
		printf("\nCompression statistic for %u compressions:\n\n"
			"  Input length ... %5u bytes   %6u bits\n"
			"   Best result ... %5u bytes   %6u bits ... ",
			compressions, input_size, input_size << 3, (best_len_result + 7) >> 3, best_len_result);
		DisplayCompression(best_len_compress); putchar('\n');
		if (depackers && best_dec_compress >= 0)
		{
			printf(
				"   With depack ... %5u bytes   %6u bits ... ",
				best_dec_result, best_dec_result<<3);
			DisplayCompression(best_dec_compress); putchar('\n');
		}

		for (int s = 1; s >= (compressions == 1 ? 1:0); s--)
		{
			puts(s ? (compressions < 2 ? "\nResults:"
				: "\nResults for all used compressions:")
					: (depackers ? "\nResults sorted by length with depacker:"
						:"\nResults sorted by compressed length:"));

			printf("\nCompression                                NumSek Packed NoPck     Overhead     Packed length "); if (depackers) printf("  With depacker");
			printf("\n========================================== ====== ====== =====  ==============  =============="); if (depackers) printf("  =============");
			putchar('\n');

			unsigned este = 1;
			unsigned disp = 0;
			while (este)
			{
				este = 0;
				unsigned pack = 999999;
				for (unsigned i = 0; i < compressions; i++)
				{
					unsigned bits = depackers ? sumdec[i] : sumall[i];
					if (s || (bits && bits == disp))
					{
						char bufove[16]; sprintf(bufove, "(%u)", sumove[i]);
						char bufsiz[16]; sprintf(bufsiz, "(%u)", sumall[i]);
						DisplayCompression(i);
						printf(" %5u  %5u %5u  %5u %8s  %5u %8s",
							sumsek[i], sumpck[i], sumunp[i], (sumove[i] + 7) >> 3, bufove, (sumall[i] + 7) >> 3, bufsiz);
						if (depackers && declen[i]) printf("  +%4u = %5u", declen[i], sumdec[i]);

						if (compressions > 1)
						{
							if ((int)sumall[i] == best_len_result &&
								(int)sumdec[i] == best_dec_result) printf(" ... best compress and depacker");
							else if ((int)sumall[i] == best_len_result) printf(" ... best compress");
							else if ((int)sumdec[i] == best_dec_result) printf(" ... best depacker");

							if ((int)sumall[i] == wrst_len_result &&
								(int)sumdec[i] == wrst_dec_result) printf(" ... worst compress and depacker");
							else if ((int)sumall[i] == wrst_len_result) printf(" ... worst compress");
							else if ((int)sumdec[i] == wrst_dec_result) printf(" ... worst depacker");
						}

						putchar('\n');
					}
					if (bits > disp && pack > bits)
					{
						pack = bits;
						este = 1;
					}
				}
				disp = pack;
				if (s) break;
			}
		}
		if (depackers && compressions > 1)
		{
			unsigned sum = 0;
			for (unsigned com = 0; com < compressions; com++) if (!declen[com]) sum++;
			if (sum) printf("\nNote: Sorted list does not contain %u compressions with not defined depacker length.\n", sum);
		}
		putchar('\n');
	}

	////////////////////////////
	// Generating output data //
	////////////////////////////

	time_stat_out = clock();

	if (!show_statistics && make_all_compressions && compressions > 1) putchar('\n');
	int best_of_best = depackers && best_dec_compress >= 0 ? best_dec_compress : best_len_compress;

	unsigned packed_size = (unsigned)-1;
	int tmp_distance = -1;
	int min_distance = -1;

	for (int comsel = (int)make_all_compressions ? 0 : best_of_best;
			 comsel <= (make_all_compressions ? (int)compressions - 1 : best_of_best);
			 comsel ++)
	{
		packed_size = (sumall[comsel] + 7) >> 3;

		if (ENAOUT)
		{
			printf("\nGenerating output data for ");
			DisplayCompression(comsel);
			printf("  Total packed length: %u bytes  %u bits\n\n", packed_size, sumall[comsel]);
		}

		tmp_distance = 0;
		min_distance = 0;

		idxout = 0;
		bits_stream = 0;
		packed_index = 0;
		packed_bytes = 0;
		unpack_bytes = 0;
		overhead_bits = 0;

		tmp_distance = packed_size - input_size + idxout - packed_index;
		if (min_distance < tmp_distance) min_distance = tmp_distance;

		if (ena_first_byte[typcom[comsel]])
		{
			if (ENAOUT && ENAZAC == 0)
			{
				printf("Data     +0 0000   Packed     +0 0000  Tax:0      ");
				printf("Dist:%6d => %-5d ", tmp_distance, min_distance);
				printf("First byte: 0x%02X\n", *input_data);
			}

			*packed_data = *input_data;
			packed_index++;
			unpack_bytes++;
			idxout++;
		}

		while (idxout < input_size)
		{
			if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
			{
				printf("Data %+6d %04X   Packed %+6d %04X  Tax:%-6u ",
					idxout, idxout, packed_index, packed_index, overhead_bits + (unpack_bytes << 3));
				printf("Dist:%6d => %-5d ", tmp_distance, min_distance);
			}

			int length = memlen[idxout][comsel];
			int offset = memofs[idxout][comsel];

			if (idxout + length < input_size && memofs[idxout + length][comsel] == 0xFFFF)
			{
				// Case for Unpacked block or copied byte and then sekvence with reused offset //
				int kdesek = idxout + length;
				int lensek = memlen[kdesek][comsel];

				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("%s Off:%-2d Len:%-5u Len: %-5u Overhead: %-2u   ",
						offset ? "Cop" : "Unp",
						offset, length, lensek, GetUnpReusedOffsetSekTax(comsel, length, offset, lensek) - (offset ? 0 : (length << 3)));

				int tmp_bits = overhead_bits;
				PutReusedOffset(comsel, idxout, length, offset, lensek);

				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("  (tax:%u dif:%+d)",
						overhead_bits - tmp_bits,
						overhead_bits - tmp_bits + (offset ? 0 : (length << 3)) - GetUnpReusedOffsetSekTax(comsel, length, offset, lensek));

				if (offset)
				{
					packed_bytes += lensek + 1;
				}
				else
				{
					unpack_bytes += length;
					packed_bytes += lensek;
				}
				idxout += length + lensek;
			}
			else if (offset)
			{
				// Case for normal sekvence with offset //
				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("Sekvence  Off: %-5u Len: %-5u Overhead: %-2u   ",
						offset, length, GetSekvenceTax(comsel, offset, length));

				int tmp_bits = overhead_bits;
				PutSekvence(comsel, offset, length);
				packed_bytes += length;
				idxout += length;

				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("  (tax:%u dif:%+d)",
						overhead_bits - tmp_bits,
						overhead_bits - tmp_bits - GetSekvenceTax(comsel, offset, length));
			}
			else
			{
				// Case for unpacked data //
				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("Unpacked  .........  Len: %-5u Overhead: %-2u   ",
						length, unptax[typcom[comsel]][length] - (length << 3));

				int tmp_bits = overhead_bits;
				PutUnpackData(comsel, idxout, length);
				unpack_bytes += length;
				idxout += length;

				if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND)
					printf("  (tax:%u dif:%+d)",
						overhead_bits - tmp_bits,
						overhead_bits - tmp_bits + (length << 3) - (int)unptax[typcom[comsel]][length]);
			}
			tmp_distance = packed_size - input_size + idxout - packed_index;
			if (min_distance < tmp_distance) min_distance = tmp_distance;

			if (ENAOUT && (int)idxout >= ENAZAC && idxout <= ENAEND) putchar('\n');
		}

		if (ENAOUT)
		{
			printf("Data %+6d %04X   Packed %+6d %04X  Tax:%-6u ",
				idxout, idxout, packed_index, packed_index, overhead_bits + (unpack_bytes << 3));
			printf("Dist:%6d => %-5d ", tmp_distance, min_distance);
			printf("End mark ");
		}

		// Put End mark //
		PutValue(endmark_pattern[typcom[comsel]], endmark_length[typcom[comsel]]);

		if (ENAOUT) putchar('\n');

		unsigned total_tax = overhead_bits + (unpack_bytes << 3);
		unsigned bits_overhead = overhead_bits;
		if (bits_stream)
		{
			if (ENAOUT)
			{
				printf("Data %+6d %04X   Packed %+6d %04X  Tax:%-6u ",
					idxout, idxout, packed_index, packed_index, overhead_bits + (unpack_bytes << 3));
				printf("Dist:%6d => %-5d ", tmp_distance, min_distance);
				if (ENAOUT) printf("Fill last byte ");
			}

			while (bits_stream) PutBit(0);

			if (ENAOUT) putchar('\n');
		}

		tmp_distance = packed_size - input_size + idxout - packed_index;
		if (min_distance < tmp_distance) min_distance = tmp_distance;

		if (ENAOUT)
		{
			printf("Data %+6d %04X   Packed %+6d %04X  Tax:%-6u ",
				idxout, idxout, packed_index, packed_index, overhead_bits + (unpack_bytes << 3));
			printf("Dist:%6d => %-5d\n\n", tmp_distance, min_distance);
		}

		if (sumall[comsel] != total_tax)
		{
			printf("Error: Bad length (bits)   Estimated: %-6u  Output: %-6u  Dif: %+-5d  in  ", sumall[comsel], total_tax, total_tax - sumall[comsel]);
			DisplayCompression(comsel);
			putchar('\n');
		}

		if (packed_size != packed_index)
		{
			printf("Error: Bad length (bytes)  Estimated: %-6u  Output: %-6u  Dif: %+-5d  in  ", packed_size, packed_index, packed_index - packed_size);
			DisplayCompression(comsel);
			putchar('\n');
		}

		packed_size = packed_index;

		if (reverse_order)
			for (int i = (packed_size >> 1) - 1; i >= 0; i--)
			{
				unsigned char tmp = packed_data[i];
				packed_data[i] = packed_data[packed_size - i - 1];
				packed_data[packed_size - i - 1] = tmp;
			}

		min_distance++;

		///////////////////////
		// Write output file //
		///////////////////////

		char *txt = buffer + nam_len;

		memcpy(buffer, nam_out, nam_len);
		if (!set_output_name || make_all_compressions) txt = GetCompressType(txt, comsel);
		strcpy(txt, ext_out);

		printf("Save file: %-*s  Length: %-5u  Packed: %-5u  NoPacked: %-5u  Overhead: %-5u (%u bits)\n",
			set_indent, buffer, packed_size, packed_bytes, unpack_bytes, (bits_overhead + 7) >> 3, bits_overhead);

		FILE *file_out = fopen(buffer, "wb");
		if (!file_out) { perror(buffer); exit(4); }
		fwrite(packed_data, 1, packed_size, file_out);
		fclose(file_out);

		if (make_include_file)
		{
			strcpy(txt, ".inc");
			file_out = fopen(buffer, "wt");
			if (!file_out) { perror(buffer); exit(4); }

			fprintf(file_out,
				"revers\t=\t%u\n"
				"typcom\t=\t%u\n"
				"typpos\t=\t%u\n"
				"bytcop\t=\t%u\n"
				"ofset1\t=\t%u\n"
				"ofset2\t=\t%u\n\n"
				"mindst\t=\t%u\n"
				"maxdct\t=\t%u\n"
				"deplen\t=\t%u\n"
				"pcklen\t=\t%u\n"
				"totlen\t=\t%u\n",
				reverse_order,
				typcom[comsel],
				typpos[comsel],
				bytcop[comsel],
				ofset1[comsel],
				ofset2[comsel],
				min_distance,
				offset_limit,
				declen[comsel],
				packed_size,
				input_size);

			fclose(file_out);
		}
	}
	putchar('\n');

	time_finish = clock();

	//////////////////////////////////////////////////////////////////////////////////////////////////////

	printf("Tested com types: %u\n", compressions);

	if (!make_all_compressions)
	{
		GetCompressType(buffer, best_of_best);
		printf("Minimal distance: %u byte%s\n", min_distance, min_distance == 1 ? "":"s");
		printf("Used compression: %s\n", buffer);
	}

	if (ENACOM) printf(" Sum of compares: %u\n", cmps);

	if (CLOCKS_PER_SEC == 1000)		// (should be defined as 1000 in time.h) //
	{
		int time_diff = time_finish - time_start;
		if (ENATIM || time_diff > 0) printf(" Processing time: %u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
		if (ENATIM)
		{
			time_diff = time_prep_lz - time_start;    printf("\n    Prepare time: %2u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
			time_diff = time_lz_opti - time_prep_lz;    printf(" LZ compres time: %2u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
			time_diff = time_opti_stat - time_lz_opti;  printf("   Optimize time: %2u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
			time_diff = time_stat_out - time_opti_stat; printf(" Statistics time: %2u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
			time_diff = time_finish - time_stat_out;    printf("  Write out time: %2u.%03u seconds\n", time_diff / 1000, time_diff % 1000);
		}
	}
	
	return 0;
}
